module calc_multigrid
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO
  use set_classification, only: ncalc, c_end, nxgrid, nygrid, calc_end,&
                                cell2calc, out_flag
  use set_calculation, only: cell_bound, bound_cflag
  use set_mg_condition, only: fine2coas, prolo_index, prolo_wei, prolo_num,&
                              restr_index, restr_wei, restr_num

  implicit none
  private
  public :: calc_interc2f, calc_prolo, calc_restr
  public :: set_boundmg, calc_errcoa, calc_correc

  ! -- local
  integer(I4) :: i, j, k, c, cc
  integer(I4) :: nc, nf, fx, fy, fz, nc_num
  real(DP) :: sum_wei, sum_x

  contains

  subroutine calc_interc2f(ng, intx)
  !*****************************************************************************
  ! calc_interc2f --  Calcultae interpolation coaser --> finer
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: ng
    real(DP), intent(inout) :: intx(:)
    ! -- local

    !---------------------------------------------------------------------------
    do nc = 1, c_end(ng) - c_end(ng-1)
      nc_num = c_end(ng-1) + nc
      if (out_flag(nc_num) /= 1) then
        cc = cell2calc(nc_num)
        if (bound_cflag(cc) /= 1) then
          if (mod(nc,nxgrid(ng)) == 0) then
            fx = nxgrid(ng)*2 - 1
          else
            fx = mod(nc,nxgrid(ng))*2 - 1
          end if

          fz = (nc-1)/(nxgrid(ng)*nygrid(ng))*2 + 1
          fy = (nc-1)/nxgrid(ng)*2 + 1 - (fz-1)*nygrid(ng)

          do k = fz, fz+1
            do j = fy, fy+1
              do i = fx, fx+1
                nf = c_end(ng-2) + nxgrid(ng-1)*nygrid(ng-1)*(k-1) + nxgrid(ng-1)*(j-1) + i
                if (out_flag(nf) /= 1) then
                  c = cell2calc(nf)
                  if (bound_cflag(c) /= 1) then
                    intx(c) = intx(cc)
                  end if
                end if
              end do
            end do
          end do
        end if
      end if
    end do

    return

  end subroutine calc_interc2f

  subroutine calc_prolo(ng, prox)
  !*****************************************************************************
  ! calc_prolo --  Calcultae prolongation coaser --> finer
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: ng
    real(DP), intent(inout) :: prox(ncalc)
    ! -- local

    !---------------------------------------------------------------------------
    do i = calc_end(ng-1)+1, calc_end(ng)
      if (bound_cflag(i) /= 1 .and. fine2coas(i) /= 0) then
        nc_num = fine2coas(i)
        sum_wei = 27.00_DP
        sum_x = prox(nc_num)*sum_wei
        do k = prolo_index(i-1)+1, prolo_index(i)
          sum_wei = sum_wei + prolo_wei(k)
          sum_x = sum_x + prox(prolo_num(k))*prolo_wei(k)
        end do
        prox(i) = sum_x/sum_wei
      end if
    end do

    return

  end subroutine calc_prolo

  subroutine calc_restr(ng, resx)
  !*****************************************************************************
  ! calc_restr --  Calculate restriction finer --> coaser
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: ng
    real(DP), intent(inout) :: resx(ncalc)
    ! -- local

    !---------------------------------------------------------------------------
    do i = calc_end(ng)+1, calc_end(ng+1)
      sum_wei = DZERO ; sum_x = DZERO
      if (bound_cflag(i) /= 1) then
        do k = restr_index(i-1)+1, restr_index(i)
          sum_wei = sum_wei + restr_wei(k)
          sum_x = sum_x + resx(restr_num(k))*restr_wei(k)
        end do
        resx(i) = sum_x/sum_wei
      else
        resx(i) = cell_bound(i)
      end if
    end do

    return

  end subroutine calc_restr

  subroutine set_boundmg(mglevel, boundx)
  !*****************************************************************************
  ! set_boundmg --  Set boundary condition at finer grid
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: mglevel
    real(DP), intent(inout) :: boundx(ncalc)
    ! -- local
    integer(I4) :: bound_sta, bound_end
    !---------------------------------------------------------------------------
    bound_sta = calc_end(mglevel-1) + 1
    bound_end = calc_end(mglevel)

    do i = bound_sta, bound_end
      if (bound_cflag(i) == 1) then
        boundx(i) = cell_bound(i)
      end if
    end do

    return

  end subroutine set_boundmg

  subroutine calc_errcoa(mglevel, restrx, coasx, coase)
  !*****************************************************************************
  ! calc_errcoa -- Calculate error on coaser grid
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: mglevel
    real(DP) , intent(in) :: restrx(:), coasx(:)
    real(DP) , intent(out) :: coase(ncalc)
    ! -- local
    integer(I4) :: err_sta, err_end
    !---------------------------------------------------------------------------
    err_sta = calc_end(mglevel) + 1
    err_end = calc_end(mglevel+1)

    do i = err_sta, err_end
      coase(i) = coasx(i) - restrx(i)
    end do

    return

  end subroutine calc_errcoa

  subroutine calc_correc(mglevel, core, corx)
  !*****************************************************************************
  ! calc_correc -- Calculate correction
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: mglevel
    real(DP), intent(in) :: core(:)
    real(DP), intent(inout) :: corx(ncalc)
    ! -- local
    integer(I4) :: cor_sta, cor_end, xi
    !---------------------------------------------------------------------------
    cor_sta = calc_end(mglevel-1) + 1
    cor_end = calc_end(mglevel)

    do i = cor_sta, cor_end
      corx(i) = corx(i) + core(i)
    end do

    return

  end subroutine calc_correc

end module calc_multigrid
