module calc_parameter
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE, DTWO, DHALF
  use set_classification, only: ncalc, calc2cell
  use make_cell, only: cell_topelev
  use set_calculation, only: cell_retn, cell_reta, cell_sat, cell_res

  implicit none
  private
  public :: calc_relperm, calc_srati, calc_hyd_geoharm, calc_hyd_upwind
  public :: calc_hyd_harm, calc_hyd_geo, calc_hyd_arith

  ! -- local
  real(DP) :: retm = DZERO, beta = DZERO

  contains

  subroutine calc_relperm(num, pertur, pres, srati, rperm)
  !*****************************************************************************
  ! calc_relperm -- Calculate relative permeability
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: num
    real(DP), intent(in) :: pertur
    real(DP), intent(in) :: pres(:)
    real(DP), intent(out) :: srati(ncalc)
    real(DP), intent(out) :: rperm(ncalc)
    ! -- local
    integer(I4) :: n
    real(DP) :: kr, se
    !---------------------------------------------------------------------------
    do n = 1, num
      ! -- Calculate saturation ratio (sratio)
        call calc_srati(n, pertur, pres(n), srati(n))

      if (srati(n) < DONE) then
!        kr = (DONE+beta)**(-2.5_DP*retm)*((DONE+beta)**retm-beta**retm)**DTWO
        se = (DONE+beta)**(-retm)
        kr = se**(DHALF)*(DONE-(DONE-se**(DONE/retm))**retm)**DTWO
      else
        kr = DONE
      end if
      rperm(n)= kr
    end do

    return

  end subroutine calc_relperm

  subroutine calc_srati(i, pertur, whead, satu_ratio)
  !*****************************************************************************
  ! calc_srati -- Calculate saturation ratio by van genuchten model
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: i
    real(DP), intent(in) :: pertur, whead
    real(DP), intent(out) :: satu_ratio
    ! -- local
    real(DP) :: per_phead, phead0, theta
    integer(I4) :: c
    !---------------------------------------------------------------------------
    c = calc2cell(i)
    per_phead = whead - cell_topelev(c) + pertur
    phead0 = DZERO
!    phead0 = -0.30

    if (per_phead < DZERO) then
      retm = DONE - DONE/cell_retn(i)
      if (per_phead <= phead0) then
        beta = abs(per_phead*cell_reta(i))**cell_retn(i)
        theta = cell_res(i) + (cell_sat(i)-cell_res(i))*((DONE+beta))**(-retm)
      else
!        beta = abs(phead0*cell_reta(i))**cell_retn(i)
        beta = DZERO
!        theta = cell_res(i) + (cell_sat(i)-cell_res(i))*((DONE+beta))**(-retm)
        theta = cell_sat(i)
      end if
      satu_ratio = theta/cell_sat(i)
    else
      satu_ratio = DONE
    end if

    return

  end subroutine calc_srati

  subroutine calc_hyd_geoharm(sratio1, sratio2, hyd_c1, hyd_c2, d1, d2, hyd_gh)
  !*****************************************************************************
  ! calc_hyd_geoharm -- Calculate hydradulic conductivity by geometric & harmonic
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: sratio1, sratio2, hyd_c1, hyd_c2, d1, d2
    real(DP), intent(out) :: hyd_gh
    ! -- local
    real(DP) :: c1, c2, d12, cd
    !---------------------------------------------------------------------------
    d12 = d1 + d2

    if (sratio1 == DONE .and. sratio2 == DONE) then !harmonic mean
      c1 = hyd_c1*d12
      c2 = hyd_c2*d12
      cd = DONE/(d1*c2+d2*c1)
      hyd_gh = (c1*c2)*cd
    else !geometric mean
      d12 = DONE/d12
      c1 = d1*log(hyd_c1)
      c2 = d2*log(hyd_c2)
      hyd_gh = exp((c1+c2)*d12)
    end if

    return

  end subroutine calc_hyd_geoharm

  subroutine calc_hyd_upwind(head1, head2, hyd_c1, hyd_c2, hyd_upwind)
  !*****************************************************************************
  ! calc_hyd_upwind -- Calculate hydradulic conductivity by upwind
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: head1, head2, hyd_c1, hyd_c2
    real(DP), intent(out) :: hyd_upwind
    ! -- local

    !---------------------------------------------------------------------------
    if (head1 > head2) then
      hyd_upwind = hyd_c1
    else
      hyd_upwind = hyd_c2
    end if

    return

  end subroutine calc_hyd_upwind

  subroutine calc_hyd_harm(hyd_c1, hyd_c2, d1, d2, hyd_harm)
  !*****************************************************************************
  ! calc_hyd_harm -- Calculate hydradulic conductivity by harmonic mean
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: hyd_c1, hyd_c2, d1, d2
    real(DP), intent(out) :: hyd_harm
    ! -- local
    real(DP) :: c1, c2, d12, cd
    !---------------------------------------------------------------------------
    d12 = d1 + d2

    c1 = hyd_c1*d12
    c2 = hyd_c2*d12
    cd = DONE/(d1*c2+d2*c1)
    hyd_harm = (c1*c2)*cd

    return

  end subroutine calc_hyd_harm

  subroutine calc_hyd_geo(hyd_c1, hyd_c2, d1, d2, hyd_geo)
  !*****************************************************************************
  ! calc_hyd_geo -- Calculate hydradulic conductivity by geometric mean
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: hyd_c1, hyd_c2, d1, d2
    real(DP), intent(out) :: hyd_geo
    ! -- local
    real(DP) :: c1, c2, d12
    !---------------------------------------------------------------------------
    d12 = d1 + d2
    d12 = DONE/d12

    c1 = d1*log(hyd_c1)
    c2 = d2*log(hyd_c2)
    hyd_geo = exp((c1+c2)*d12)

    return

  end subroutine calc_hyd_geo

  subroutine calc_hyd_arith(hyd_c1, hyd_c2, d1, d2, hyd_arith)
  !*****************************************************************************
  ! calc_hyd_arith -- Calculate hydradulic conductivity by arithmetic mean
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: hyd_c1, hyd_c2, d1, d2
    real(DP), intent(out) :: hyd_arith
    ! -- local
    real(DP) :: c1, c2, d12
    !---------------------------------------------------------------------------
    d12 = d1 + d2
    d12 = DONE/d12

    c1 = d1*hyd_c1
    c2 = d2*hyd_c2
    hyd_arith = (c1+c2)*d12

    return

  end subroutine calc_hyd_arith

end module calc_parameter
