module linear_solution
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE, DINFI
  use read_inputfile, only: solver_flag, maxinn_iter, sor_omega, sip_alpha,&
                            nlevel, maxvcy_iter, max_swp, errtol
  use set_classification, only: ncalc, calc_end
  use set_condition, only: off_index, off_row, conn_dir
  use set_mg_condition, only: set_mgvolave
  use set_calculation, only: bound_cflag
  use calc_multigrid, only: calc_interc2f, calc_prolo, calc_restr, set_boundmg,&
                            calc_errcoa, calc_correc
  use allocate_module, only: dmat, lumat, rhs, resi
  use time_module, only: conv_flag, form_swich
  use check_simulation, only: check_insol
  use calc_simulation, only: calc_l2norm2, calc_resi, calc_resl2norm
  use calc_function, only: calc_func, calc_vecjacf

  implicit none
  private
  public :: solve_linalg
  integer(I4), public :: iter

  ! -- local
  integer(I4) :: n ,i, k, kk, st_num, en_num, nesiter, nlev
  real(DP) :: s, bnorm, rnorm

  contains

  subroutine solve_linalg(x, init_norm)
  !*****************************************************************************
  ! solve_lnralg -- Solve linear algebra
  !*****************************************************************************
    ! -- module

    ! -- inout
    real(DP), intent(inout) :: x(:)
    real(DP), intent(in) :: init_norm
!    real(DP), intent(out) :: last_norm
    ! -- local

    !---------------------------------------------------------------------------
    bnorm = init_norm

    if (solver_flag < 5 .or. solver_flag == 9) then
      st_num = calc_end(0) + 1
      en_num = calc_end(1)
      nesiter = 1
      if (solver_flag == 1) then
      ! -- Solve Gauss-seidel method (gs)
        call solve_gs(1, x, rhs)
      else if (solver_flag == 2) then
      ! -- Solve Successive Over Relaxation method (sor)
        call solve_sor(1, x, rhs)
      else if (solver_flag == 3) then
      ! -- Solve Strongly Implicit Procedure method (sip)
        call solve_sip(1, x, rhs)
      else if (solver_flag == 4 .or. solver_flag == 9) then
        if (form_swich == 0) then
        ! -- Solve Preconditioned Conjugate Gradient method (pcg)
          call solve_pcg(1, x, rhs)
        else if (form_swich == 1) then
        ! -- Solve Preconditioned Bi-Conjugate Gradient Stabilized method (bicgs)
          call solve_bicgs(1, x, rhs)
        end if
        if (solver_flag == 9) then
          do nlev = 1, nlevel-1
            ! -- Calculate restriction finer --> coaser (restr)
              call calc_restr(nlev, x)
          end do
        end if
      end if
    else
    ! -- Solve Multi Grid Vcycle (mgv)
      call solve_mgv(x)
!    ! -- Solve Full Multi Grid (fmg)
!      call solve_fmg(x)
    end if

!    last_norm = rnorm

    return

  end subroutine solve_linalg

  subroutine solve_gs(level, inx, inb)
  !*****************************************************************************
  ! solve_gs -- Solve Gauss-seidel method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(st_num:en_num), inb(st_num:en_num)
    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calculate l2norm square (l2norm2)
      call calc_l2norm2(level, inb, bnorm)

    do iter = 1, maxinn_iter
      do n = st_num, en_num
        s = DZERO
        do k = off_index(n-1)+1, off_index(n)
          s = s + lumat(k)*inx(off_row(k))
        end do
        inx(n) = DONE/dmat(n)*(inb(n) - s)
      end do

      if (solver_flag < 5) then
        ! -- Calculate residual &l2norm square (resl2norm)
          call calc_resl2norm(st_num, en_num, inx, inb, rnorm)
        ! -- Check inner solution (check_insol)
          call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        end if
      else
        exit
      end if

      if (iter == maxinn_iter) then
        exit
      end if

    end do

    return

  end subroutine solve_gs

  subroutine smooth_gs(level, inx, inb)
  !*****************************************************************************
  ! smooth_gs -- Smooth by Gauss-seidel method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(ncalc), inb(ncalc)
    ! -- local
    integer(I4) :: ns
    !---------------------------------------------------------------------------
    if (level == nlevel) then
      ns = maxinn_iter
    else
      ns = 1
    end if

    do i = 1, ns
      do n = st_num, en_num
        s = DZERO
        do k = off_index(n-1)+1, off_index(n)
          s = s + lumat(k)*inx(off_row(k))
        end do
        inx(n) = DONE/dmat(n)*(inb(n) - s)
      end do
    end do

    return

  end subroutine smooth_gs

  subroutine solve_sor(level, inx, inb)
  !*****************************************************************************
  ! solve_sor -- Solve Successive Over Relaxation method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(st_num:en_num), inb(st_num:en_num)
    ! -- local
    real(DP), allocatable :: x_old(:)
    !---------------------------------------------------------------------------
    allocate(x_old(st_num:en_num))
    x_old(:) = DZERO

    ! -- Calculate l2norm square (l2norm2)
      call calc_l2norm2(level, inb, bnorm)

    do iter = 1, maxinn_iter
      x_old(st_num:en_num) = inx(st_num:en_num)
      do n = st_num, en_num
        s = DZERO
        do k = off_index(n-1)+1, off_index(n)
          s = s + lumat(k)*inx(off_row(k))
        end do
        inx(n) = DONE/dmat(n)*(inb(n) - s)
        inx(n) = sor_omega*inx(n) + (DONE - sor_omega)*(x_old(n))
      end do

      if (solver_flag < 5 .or. nesiter == level) then
        ! -- Calculate residual &l2norm square (resl2norm)
          call calc_resl2norm(st_num, en_num, inx, inb, rnorm)
        ! -- Check inner solution (check_insol)
          call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        end if
      else if (nesiter /= level) then
        exit
      end if

      if (iter == maxinn_iter) then
        exit
      end if

    end do

    deallocate(x_old)

    return

  end subroutine solve_sor

  subroutine solve_sip(level, inx, inb)
  !*****************************************************************************
  ! solve_sip -- Solve Strongly Implicit Procedure method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(st_num:en_num), inb(st_num:en_num)
    ! -- local
    real(DP) :: at(ncalc), as(ncalc), aw(ncalc), ae(ncalc), an(ncalc), ab(ncalc)
    integer(I4) :: dirt(ncalc), dirs(ncalc), dirw(ncalc), dire(ncalc), dirn(ncalc), dirb(ncalc)
    real(DP) :: lt(ncalc), ls(ncalc), lw(ncalc), d(ncalc), res(ncalc)
    real(DP) :: ue(0:ncalc), un(0:ncalc), ub(0:ncalc)
    real(DP) :: p1e, p1n, p1b, p2
    real(DP) :: tx, sx, wx, ex, nx, bx, tr, sr, wr, er, nr, br
    !---------------------------------------------------------------------------
    at(:) = DZERO ; as(:) = DZERO ; aw(:) = DZERO ; ae(:) = DZERO ;&
    an(:) = DZERO ; ab(:) = DZERO
    dirt(:) = 0 ; dirs(:) = 0 ; dirw(:) = 0 ; dire(:) = 0 ; dirn(:) = 0 ; dirb(:) = 0
    lt(:) = DZERO ; ls(:) = DZERO ; lw(:) = DZERO ; ue(:) = DZERO ;&
    un(:) = DZERO ; ub(:) = DZERO ; d(:) = DZERO ; res(:) = DZERO

    ! -- Calculate l2norm square (l2norm2)
      call calc_l2norm2(level, inb, bnorm)

    do n = st_num, en_num
      do k = off_index(n-1)+1, off_index(n)
        if (conn_dir(k) == 1) then
          at(n) = lumat(k)
          dirt(n) = off_row(k)
        else if (conn_dir(k) == 2) then
          as(n) = lumat(k)
          dirs(n) = off_row(k)
        else if (conn_dir(k) == 3) then
          aw(n) = lumat(k)
          dirw(n) = off_row(k)
        else if (conn_dir(k) == 4) then
          ae(n) = lumat(k)
          dire(n) = off_row(k)
        else if (conn_dir(k) == 5) then
          an(n) = lumat(k)
          dirn(n) = off_row(k)
        else if (conn_dir(k) == 6) then
          ab(n) = lumat(k)
          dirb(n) = off_row(k)
        end if
      end do

      lt(n) = at(n)/(DONE + sip_alpha*(ue(dirt(n)) + un(dirt(n))))
      ls(n) = as(n)/(DONE + sip_alpha*(ue(dirs(n)) + ub(dirs(n))))
      lw(n) = aw(n)/(DONE + sip_alpha*(un(dirw(n)) + ub(dirw(n))))
      p1e = sip_alpha*(lt(n)*ue(dirt(n)) + ls(n)*ue(dirs(n)))
      p1n = sip_alpha*(lt(n)*un(dirt(n)) + lw(n)*un(dirw(n)))
      p1b = sip_alpha*(ls(n)*ub(dirs(n)) + lw(n)*ub(dirw(n)))
      p2 = lt(n)*ub(dirt(n)) + ls(n)*un(dirs(n)) + lw(n)*ue(dire(n))
      d(n) = DONE/(dmat(n) + p1e + p1n + p1b - p2)
      ue(n) = (ae(n) - p1e)*d(n)
      un(n) = (an(n) - p1n)*d(n)
      ub(n) = (ab(n) - p1b)*d(n)
    end do

    do iter = 1, maxinn_iter
      do n = st_num, en_num
        if (dirt(n) == 0) then
          tx = DZERO ; tr = DZERO
        else
          tx = at(n)*inx(dirt(n)) ; tr = lt(n)*res(dirt(n))
        end if
        if (dirs(n) == 0) then
          sx = DZERO ; sr = DZERO
        else
          sx = as(n)*inx(dirs(n)) ; sr = ls(n)*res(dirs(n))
        end if
        if (dirw(n) == 0) then
          wx = DZERO ; wr = DZERO
        else
          wx = aw(n)*inx(dirw(n)) ; wr = lw(n)*res(dirw(n))
        end if
        if (dire(n) == 0) then
          ex = DZERO ; er = DZERO
        else
          ex = ae(n)*inx(dire(n)) ; er = ue(n)*res(dire(n))
        end if
        if (dirn(n) == 0) then
          nx = DZERO ; nr = DZERO
        else
          nx = an(n)*inx(dirn(n)) ; nr = un(n)*res(dirn(n))
        end if
        if (dirb(n) == 0) then
          bx = DZERO ; br = DZERO
        else
          bx = ab(n)*inx(dirb(n)) ; br = ub(n)*res(dirb(n))
        end if

        res(n) = inb(n) - tx - sx - wx - ex - nx - bx - dmat(n)*inx(n)
        res(n) = (res(n)-tr-sr-wr)*d(n)
      end do

      do n = en_num, st_num, -1
        res(n) = res(n) - er - nr - br
        inx(n) = inx(n) + res(n)
      end do

      if (solver_flag < 5 .or. nesiter == level) then
        ! -- Calculate residual &l2norm square (resl2norm)
          call calc_resl2norm(st_num, en_num, inx, inb, rnorm)
        ! -- Check inner solution (check_insol)
          call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        end if
      else if (nesiter /= level) then
        exit
      end if

      if (iter == maxinn_iter) then
        exit
      end if

    end do

    return

  end subroutine solve_sip

  subroutine solve_pcg(level, inx, inb)
  !*****************************************************************************
  ! solve_pcg -- Solve Preconditioned Conjugate Gradient method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(:), inb(:)
    ! -- local
    real(DP) :: sk, sk0, sk2, alpha, beta
    real(DP) :: d(ncalc), z(ncalc), p(ncalc), q(ncalc)
    !---------------------------------------------------------------------------
    ! initail setting
    sk = DZERO ; sk0 = DINFI ; sk2 = DZERO ; alpha = DZERO ; beta = DZERO
    d(:) = DZERO ; z(:) = DZERO ; p(:) = DZERO ; q(:) = DZERO

    if (solver_flag == 4 .or. solver_flag == 8) then
      ! -- Preconditon Incomplete LU diagonal (dilu)
        call precon_dilu(level, d)
    end if

    if (solver_flag == 4 .or. solver_flag == 9) then
    ! -- Calculate residual (resi)
      call calc_resi(level, inx, inb, resi)
    else if (solver_flag == 8) then
      resi(st_num:en_num) = inb(st_num:en_num)
    end if
!    ! -- Calculate l2norm square (l2norm2)
!      call calc_l2norm2(level, resi, bnorm)

    do iter = 1, maxinn_iter
      if (solver_flag == 4 .or. solver_flag == 8) then
        ! -- Solve preconditioned linear system (presys)
          call solve_presys(level, resi, d, z)
      else if (solver_flag == 9) then
        ! -- Loop cycle in Correction Storage (cs)
          call loop_cs(level, z, resi)
      end if

      sk = dot_product(resi(st_num:en_num), z(st_num:en_num))

      if (iter == 1) then
        do n = st_num, en_num
          p(n) = z(n)
        end do
      else
        beta = sk/sk0
        do n = st_num, en_num
          p(n) = z(n) + beta*p(n)
        end do
      end if

      ! -- Calculate matrix-vector multiplication (matvec)
        call calc_matvec(level, p, dmat, q)
!      ! -- Calculate vector by jacobi-free (vecjocf)
!        call calc_vecjacf(level, inx, p, q)

      sk2 = dot_product(p(st_num:en_num), q(st_num:en_num))

      alpha = sk/sk2
      do n = st_num, en_num
        inx(n) = inx(n) + alpha*p(n)
        resi(n) = resi(n) - alpha*q(n)
      end do

      if (solver_flag < 5 .or. solver_flag == 9 .or. level == 1) then
        ! -- Calculate l2norm square (l2norm2)
          call calc_l2norm2(nesiter, resi, rnorm)
        ! -- Check inner solution (check_insol)
          call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        else
          sk0 = sk
        end if
      else
        exit
      end if

      if (iter == maxinn_iter) then
        exit
      end if

    end do

    return

  end subroutine solve_pcg

  subroutine solve_bicgs(level, inx, inb)
  !*****************************************************************************
  ! solve_bicgs -- Solve Preconditioned Bi-Conjugate Gradient Stabilized method
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: level
    real(DP), intent(inout) :: inx(:), inb(:)
    ! -- local
    real(DP) :: sk, sk0, sk2, alpha, beta, bicgs_omega, ts, tt
    real(DP) :: d(ncalc), z(ncalc), p(ncalc), v(ncalc), rs(ncalc), t(ncalc)
    !---------------------------------------------------------------------------
    ! initail setting
    sk = DZERO ; sk0 = DINFI ; sk2 = DZERO ; alpha = DZERO ; beta = DZERO
    bicgs_omega = DONE ; ts = DZERO ; tt = DZERO
    d(:) = DZERO ; z(:) = DZERO ; p(:) = DZERO ; v(:) = DZERO ; rs(:) = DZERO
    t(:) = DZERO

    if (solver_flag == 4 .or. solver_flag == 8) then
      ! -- Preconditon Incomplete LU diagonal (dilu)
        call precon_dilu(level, d)
    end if

    if (solver_flag == 4 .or. solver_flag == 9) then
    ! -- Calculate residual (resi)
      call calc_resi(level, inx, inb, resi)
    else if (solver_flag == 8) then
      resi(st_num:en_num) = inb(st_num:en_num)
    end if
!    ! -- Calculate l2norm square (l2norm2)
!      call calc_l2norm2(level, resi, bnorm)

    do n = st_num, en_num
      rs(n) = resi(n)
    end do

    do iter = 1, maxinn_iter
      sk = dot_product(resi(st_num:en_num), rs(st_num:en_num))

      if (sk == DZERO) then
        exit
      end if

      if (iter == 1) then
        do n = st_num, en_num
          p(n) = resi(n)
        end do
      else
        beta = (sk/sk0)*(alpha/bicgs_omega)
        do n = st_num, en_num
          p(n) = resi(n) + beta*(p(n)-bicgs_omega*v(n))
        end do
      end if

      if (solver_flag == 4 .or. solver_flag == 8) then
        ! -- Solve preconditioned linear system (presys)
          call solve_presys(level, p, d, z)
      else if (solver_flag == 9) then
        ! -- Loop cycle in Correction Storage (cs)
          call loop_cs(level, z, p)
      end if

      ! -- Calculate matrix-vector multiplication (matvec)
        call calc_matvec(level, z, dmat, v)
!      ! -- Calculate vector by jacobi-free (vecjocf)
!        call calc_vecjacf(level, inx, z, v)

      sk2 = dot_product(v(st_num:en_num), rs(st_num:en_num))

      alpha = sk/sk2

      do n = st_num, en_num
        inx(n) = inx(n) + alpha*z(n)
        resi(n) = resi(n) - alpha*v(n)
      end do

!      if (solver_flag < 5) then
      ! -- Calculate l2norm square (l2norm2)
        call calc_l2norm2(level, resi, rnorm)
      ! -- Check inner solution (check_insol)
        call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        end if
!      end if

      if (conv_flag /= 1) then
        if (solver_flag == 4 .or. solver_flag == 8) then
          ! -- Solve preconditioned linear system (presys)
            call solve_presys(level, resi, d, z)
        else if (solver_flag == 9) then
          ! -- Loop cycle in Correction Storage (cs)
            call loop_cs(level, z, resi)
        end if

      ! -- Calculate matrix-vector multiplication (matvec)
        call calc_matvec(level, z, dmat, t)
!      ! -- Calculate vector by jacobi-free (vecjocf)
!        call calc_vecjacf(level, inx, z, t)

        ts = dot_product(t(st_num:en_num), resi(st_num:en_num))
        tt = dot_product(t(st_num:en_num), t(st_num:en_num))

        bicgs_omega = ts/tt

        do n = st_num, en_num
          inx(n) = inx(n) + bicgs_omega*z(n)
          resi(n) = resi(n) - bicgs_omega*t(n)
        end do

        if (solver_flag < 5 .or. solver_flag == 9 .or. level == 1) then
        ! -- Calculate l2norm square (l2norm2)
          call calc_l2norm2(level, resi, rnorm)
        ! -- Check inner solution (check_insol)
          call check_insol(bnorm, rnorm)
          if (conv_flag == 1) then
            exit
          else
            sk0 = sk
          end if
        else
          exit
        end if
      end if

      if (iter == maxinn_iter) then
        exit
      end if

    end do

    return

  end subroutine solve_bicgs

  subroutine solve_mgv(inx)
  !*****************************************************************************
  ! solve_mgv -- Solve Multi Grid Vcycle
  !*****************************************************************************
    ! -- module

    ! -- inout
    real(DP), intent(inout) :: inx(:)
    ! -- local

    !---------------------------------------------------------------------------
    nesiter = 1
    st_num = calc_end(nesiter-1) + 1 ; en_num = calc_end(nesiter)
!    ! -- Calculate residual &l2norm square (resl2norm)
!      call calc_resl2norm(st_num, en_num, inx, rhs, bnorm)

    do iter = 1, maxinn_iter
    ! -- Loop cycle in Correction Storage (cs)
      call loop_cs(nesiter, inx, rhs)
    ! -- Calculate residual &l2norm square (resl2norm)
      call calc_resl2norm(st_num, en_num, inx, rhs, rnorm)
    ! -- Check inner solution (check_insol)
      call check_insol(bnorm, rnorm)
      if (conv_flag == 1) then
        do nlev = 1, nlevel-1
        ! -- Calculate restriction finer --> coaser (restr)
          call calc_restr(nlev, inx)
        end do
        exit
      end if

      if (iter == maxinn_iter) then
        exit
      end if
    end do

    return

  end subroutine solve_mgv

  subroutine solve_fmg(inx)
  !*****************************************************************************
  ! solve_fmg -- Solve Full Multi Grid
  !*****************************************************************************
    ! -- module

    ! -- inout
    real(DP), intent(inout) :: inx(:)
    ! -- local

    !---------------------------------------------------------------------------
    do nesiter = nlevel, 1, -1  ! nested iteration loop
      st_num = calc_end(nesiter-1) + 1 ; en_num = calc_end(nesiter)
!      ! -- Loop cycle in Correction Storage (cs)
!        call loop_cs(nesiter, inx(st_num:en_num), rhs(st_num:en_num))
        if (solver_flag == 5) then
          call solve_gs(nesiter, inx, rhs)
        else if (solver_flag == 6) then
          call solve_sor(nesiter, inx, rhs)
        else if (solver_flag == 7) then
          call solve_sip(nesiter, inx, rhs)
        else if (solver_flag == 8) then
          if (form_swich == 0) then
            call solve_pcg(nesiter, inx, rhs)
          else if (form_swich == 1) then
            call solve_bicgs(nesiter, inx, rhs)
          end if
        end if

      if (nesiter /= 1) then
!        if (conv_flag == 1) then
          ! -- Calcultae prolongation coaser --> finer (prolo)
            call calc_prolo(nesiter-1, inx)
!          ! -- Calcultae interpolation coaser --> finer (interc2f)
!            call calc_interc2f(nesiter, inx)
          ! -- Set boundary condition at finer grid (boundmg)
            call set_boundmg(nesiter-1, inx)
!        else
!          exit
!        end if
      else
!        ! -- Calculate residual &l2norm square (resl2norm)
!          call calc_resl2norm(st_num, en_num, inx, rhs, rnorm)
!        ! -- Calculate l2norm square (l2norm2)
!          call calc_l2norm2(nesiter, resi, rnorm)
!        ! -- Check inner solution (check_insol)
!          call check_insol(bnorm, rnorm)
        if (conv_flag == 1) then
          exit
        end if
      end if

    end do

    do nlev = 1, nlevel-1
      ! -- Calculate restriction finer --> coaser (restr)
        call calc_restr(nlev, inx)
    end do

    return

  end subroutine solve_fmg

  recursive subroutine loop_cs(llevel, e, r)
  !*****************************************************************************
  ! loop_cs -- Loop cycle in Correction Storage
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: llevel
    real(DP), intent(inout) :: e(ncalc), r(ncalc)
    ! -- local
    integer(I4) :: v_iter, sweep
    real(DP) :: tempe(ncalc), tempr(ncalc)
    !---------------------------------------------------------------------------
    tempe(:) = DZERO ; tempr(:) = DZERO
!    if (llevel /= nesiter) then
!      e(st_num:en_num) = DZERO
!    end if

    do v_iter = 1, maxvcy_iter  ! V-cycle loop
      if (llevel == nlevel) then
        ! solve on coarsest grid with iterative method
!        if (solver_flag == 5) then
!          call solve_gs(llevel, e, r)
!        else if (solver_flag == 6) then
!          call solve_sor(llevel, e, r)
!        else if (solver_flag == 7) then
!          call solve_sip(llevel, e, r)
!        else if (solver_flag == 8) then
!          if (form_swich == 0) then
!            call solve_pcg(llevel, e, r)
!          else if (form_swich == 1) then
!            call solve_bicgs(llevel, e, r)
!          end if
!        else if (solver_flag == 9) then
          call solve_presys(llevel, r, dmat, e)
!          call smooth_gs(llevel, e, r)
!        end if
        tempe(st_num:en_num) = e(st_num:en_num)

      else
        do sweep = 1, max_swp
          ! pre-smoothing
!          if (solver_flag == 5) then
!            call solve_gs(llevel, e, r)
!          else if (solver_flag == 6) then
!            call solve_sor(llevel, e, r)
!          else if (solver_flag == 7) then
!            call solve_sip(llevel, e, r)
!          else if (solver_flag == 8) then
!            if (form_swich == 0) then
!              call solve_pcg(llevel, e, r)
!            else if (form_swich == 1) then
!              call solve_bicgs(llevel, e, r)
!            end if
!          else if (solver_flag == 9) then
            call smooth_gs(llevel, e, r)
!          end if
        end do

        ! -- Calculate residual
          call calc_resi(llevel, e, r, tempr)
        ! -- Calculate restriction of residual finer --> coaser (restr)
          call calc_restr(llevel, tempr)
        st_num = calc_end(llevel) + 1 ; en_num = calc_end(llevel+1)
        e(st_num:en_num) = DZERO ; r(st_num:en_num) = tempr(st_num:en_num)

        ! -- Loop cycle in Correction Storage (cs)
          call loop_cs(llevel+1, e, r)
        tempe(st_num:en_num) = e(st_num:en_num)

        ! -- Calcultae prolongation coaser --> finer (prolo)
          call calc_prolo(llevel, tempe)
        ! -- Calculate correction (correc)
          call calc_correc(llevel, tempe, e)

        st_num = calc_end(llevel-1) + 1 ; en_num = calc_end(llevel)

        do sweep = 1, max_swp
          ! post-smoothing
!          if (solver_flag == 5) then
!            call solve_gs(llevel, e, r)
!          else if (solver_flag == 6) then
!            call solve_sor(llevel, e, r)
!          else if (solver_flag == 7) then
!            call solve_sip(llevel, e, r)
!          else if (solver_flag == 8) then
!            if (form_swich == 0) then
!              call solve_pcg(llevel, e, r)
!            else if (form_swich == 1) then
!              call solve_bicgs(llevel, e, r)
!            end if
!          else if (solver_flag == 9) then
            call smooth_gs(llevel, e, r)
!          end if
        end do
      end if
    end do

    return

  end subroutine loop_cs

  subroutine precon_dilu(plevel, pre_d)
  !*****************************************************************************
  ! precon_dilu -- Preconditon Incomplete LU diagonal
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: plevel
    real(DP), intent(out) :: pre_d(:)
    ! -- local
    integer(I4) :: pre_sta, pre_end
    real(DP) :: d_invk
    !---------------------------------------------------------------------------
    pre_sta = calc_end(plevel-1) + 1
    pre_end = calc_end(plevel)

    do i = pre_sta, pre_end
      pre_d(i) = dmat(i)
    end do

    do i = pre_sta, pre_end
      do k = off_index(i-1)+1, off_index(i)
        if (i > off_row(k)) then
          d_invk = DONE/pre_d(off_row(k))
          do kk = off_index(off_row(k)-1)+1, off_index(off_row(k))
            if (off_row(k) < off_row(kk)) then
              if (off_row(kk) == i) then
                pre_d(i) = pre_d(i) - lumat(k)*d_invk*lumat(kk)
              end if
            end if
          end do
        end if
      end do
    end do

    return

  end subroutine precon_dilu

  subroutine solve_presys(plevel, inrhs, indmat, outx)
  !*****************************************************************************
  ! solve_presys -- Solve preconditioned linear system
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: plevel
    real(DP), intent(in) :: inrhs(:), indmat(:)
    real(DP), intent(out) :: outx(:)
    ! -- local
    integer(I4) :: pre_sta, pre_end
    !---------------------------------------------------------------------------
    pre_sta = calc_end(plevel-1) + 1
    pre_end = calc_end(plevel)

    do i = pre_sta, pre_end
      outx(i) = inrhs(i)
    end do

    ! Forward Substitution
    do i = pre_sta, pre_end
      s = outx(i)
      do k = off_index(i-1)+1, off_index(i)
        if (i > off_row(k)) then
          s = s - lumat(k)*outx(off_row(k))
        end if
      end do
      outx(i) = s/indmat(i)
    end do

    ! Backward Substitution
    do i = pre_end, pre_sta, -1
      s = DZERO
      do k = off_index(i-1)+1, off_index(i)
        if (i < off_row(k)) then
          s = s + lumat(k)*outx(off_row(k))
        end if
      end do
      outx(i) = outx(i) - s/indmat(i)
    end do

    return

  end subroutine solve_presys

  subroutine calc_matvec(mvlevel, invec, indmat, outvec)
  !*****************************************************************************
  ! calc_matvec -- Calculate matrix-vector multiplication
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: mvlevel
    real(DP), intent(in) :: invec(:), indmat(:)
    real(DP), intent(out) :: outvec(:)
    ! -- local
    integer(I4) :: mv_sta, mv_end
    !---------------------------------------------------------------------------
    mv_sta = calc_end(mvlevel-1) + 1
    mv_end = calc_end(mvlevel)

    do i = mv_sta, mv_end
      s = indmat(i)*invec(i)
      do k = off_index(i-1)+1, off_index(i)
        s = s + lumat(k)*invec(off_row(k))
      end do
      outvec(i) = s
    end do

    return

  end subroutine calc_matvec

  subroutine set_rhscoa(mglevel, coax, coar, coarhs)
  !*****************************************************************************
  ! set_rhscoa --  Set right hand side on coaser grid
  !*****************************************************************************
    ! -- module

    ! -- inout
    integer(I4), intent(in) :: mglevel
    real(DP), intent(in) :: coax(:)
    real(DP), intent(in) :: coar(:)
    real(DP), intent(out) :: coarhs(:)
    ! -- local
    integer(I4) :: coa_sta, coa_end
    real(DP) :: coa_axsum
    !---------------------------------------------------------------------------
    coa_sta = calc_end(mglevel) + 1
    coa_end = calc_end(mglevel+1)

    do i = coa_sta, coa_end
      coa_axsum = DZERO
      coa_axsum = dmat(i)*coax(i)
      do k = off_index(i-1)+1, off_index(i)
        coa_axsum = coa_axsum + lumat(k)*coax(off_row(k))
      end do
      coarhs(i) = coar(i) + coa_axsum
    end do

    return

  end subroutine set_rhscoa

end module linear_solution
