module make_linearsystem
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE
  use read_inputfile, only: sim_flag, newper, newper_inv
  use set_classification, only: ncalc, calc2cell, calc2cals, cals2surf
  use make_cell, only: area, vol_c, dis, surfelev
  use set_condition, only: off_index, off_row, conn_dir, g_dir, g_index
  use set_calculation, only: bound_cflag, rech_cflag, well_cflag, cell_ss,&
                             cell_sat, ghost_head, surf_head
  use calc_parameter, only: calc_relperm, calc_srati, calc_hyd_geoharm,&
                            calc_hyd_upwind, calc_hyd_harm, calc_hyd_geo,&
                            calc_hyd_arith
  use allocate_module, only: head_old, head_new, hydcon_new, dmat, lumat, rhs,&
                             srati_old, srati_new, pertur_hyd, pertur_srati,&
                             calc_rech, calc_well, rel_perm, pertur_rel
  use time_module, only: delt, form_swich

  implicit none
  private
  public :: make_matvec

  ! -- local
  integer(I4) :: i, j, k, s, c, g, dir, ori
  real(DP) :: ch_stor, deri_stor, deri_srati, ch_stor1, ch_stor2
  real(DP) :: conn_flow, deri_dconn
  real(DP) :: delhead, abyd
  real(DP) :: deri_kphi1, deri_kphi2, delt_inv

  contains

  subroutine make_matvec()
  !*****************************************************************************
  ! make_matvec -- Make matrix & vector
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Calculate relative permeability (relperm)
      call calc_relperm(ncalc, DZERO, head_new, srati_new, rel_perm)
      call calc_relperm(ncalc, newper, head_new, pertur_srati, pertur_rel)

    if (sim_flag >= 0) then
      delt_inv = DONE/delt
      ! -- Form storage change (stochn)
        call form_stochn()
    end if

    ! -- Form connect flow from adjacent cells (connflow)
      call form_connflow()

    ! -- Form boundary term (boundterm)
      call form_boundterm()

    return

  end subroutine make_matvec

  subroutine form_stochn()
  !*****************************************************************************
  ! form_stochn -- Form storage change
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: alp
    !---------------------------------------------------------------------------
    do i = 1, ncalc
      deri_stor = DZERO
      deri_srati = DZERO
      if (bound_cflag(i) /= 1) then
        c = calc2cell(i)
        ch_stor1 = srati_old(i)*head_old(i)
        ch_stor2 = srati_new(i)*head_new(i)
        if (srati_new(i) < DONE) then
          alp = DZERO
        else
          alp = DONE
        end if
        ch_stor = (alp*cell_ss(i)*(ch_stor2-ch_stor1) + &
                  cell_sat(i)*(srati_new(i)-srati_old(i)))*vol_c(c)

        deri_srati = (pertur_srati(i)-srati_new(i))*newper_inv
        deri_stor = (cell_sat(i)*deri_srati+alp*cell_ss(i)*srati_new(i))*vol_c(c)

        if (form_swich == 1) then
          deri_stor = deri_stor + alp*cell_ss(i)*deri_srati*head_new(i)*vol_c(c)
        end if
        rhs(i) = rhs(i) + ch_stor*delt_inv
        dmat(i) = dmat(i) - deri_stor*delt_inv
      end if
    end do

    return

  end subroutine form_stochn

  subroutine form_connflow()
  !*****************************************************************************
  ! form_connflow -- Form connect flow from adjacent cells
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: dis1, dis2, dis12, conn_hyd, pertur_head
    real(DP) :: deri_hyd1, deri_hyd2, deri_conn1, deri_conn2
    real(DP) :: relat, hyd_face1, hyd_face2, hyd_face
    !---------------------------------------------------------------------------
    do i = 1, ncalc
      conn_flow = DZERO
      deri_dconn = DZERO
      c = calc2cell(i)
      do k = off_index(i-1)+1, off_index(i)
        dir = conn_dir(k)
        dis1 = dis(c,dir)
        j = off_row(k)
        delhead = head_new(j) - head_new(i)

        if (bound_cflag(j) == 1) then
          dis2 = dis1
          dis12 = dis1
          j = i
        else
          dis2 = dis(calc2cell(j),(7-dir))
          dis12 = dis1 + dis2
        end if
        abyd = area(c,dir)/dis12

        if (dir >= 4) then
          ori = 7 - dir
        else
          ori = dir
        end if

        ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
          call calc_hyd_upwind(head_new(i), head_new(j), rel_perm(i), rel_perm(j), relat)

        hyd_face1 = hydcon_new(i,ori)
        hyd_face2 = hydcon_new(j,ori)

        ! -- Calculate hydradulic conductivity by harmonic mean (hyd_harm)
          call calc_hyd_harm(hyd_face1, hyd_face2, dis1, dis2, hyd_face)

        conn_hyd = hyd_face*relat

        conn_flow = conn_flow + conn_hyd*abyd*delhead
        deri_conn1 = -conn_hyd*abyd
        deri_conn2 = conn_hyd*abyd

        if (form_swich == 1 .and. sim_flag >= 0) then
          pertur_head = head_new(i) + newper
          ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
            call calc_hyd_upwind(pertur_head, head_new(j), pertur_rel(i), rel_perm(j), relat)

          deri_hyd1 = hyd_face*relat

          pertur_head = head_new(j) + newper
          ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
            call calc_hyd_upwind(head_new(i), pertur_head, rel_perm(i), pertur_rel(j), relat)

          deri_hyd2 = hyd_face*relat

          deri_kphi1 = (deri_hyd1-conn_hyd)*newper_inv
          deri_kphi2 = (deri_hyd2-conn_hyd)*newper_inv
          deri_conn1 = deri_conn1 + deri_kphi1*delhead*abyd
          deri_conn2 = deri_conn2 + deri_kphi2*delhead*abyd
        end if

        deri_dconn = deri_dconn + deri_conn1

        j = off_row(k)

        lumat(k) = deri_conn2
      end do
      rhs(i) = rhs(i) - conn_flow
      dmat(i) = dmat(i) + deri_dconn
    end do

    return

  end subroutine form_connflow

  subroutine form_boundterm()
  !*****************************************************************************
  ! form_boundterm -- Form boundary term
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: ss
    !---------------------------------------------------------------------------
    do i = 1, ncalc
      if (bound_cflag(i) /= 1) then
        s = calc2cals(i)
        c = calc2cell(i)
        ! neuman boundary condition & well uptake
        if (s /= 0) then
          if (rech_cflag(s) == 1) then
            rhs(i) = rhs(i) - calc_rech(s)*area(c,1)
          end if
        end if
        if (well_cflag(i) == 1) then
          rhs(i) = rhs(i) - calc_well(i)
        end if

        ! dirichlet bounday condtion except sea level (surface water & MG ghost)
        if (bound_cflag(i) == 2) then !MG ghost
          do g = g_index(i-1)+1, g_index(i)
            dir = g_dir(g)
            delhead = ghost_head(g) - head_new(i)
            ! -- Set rhs & dmat for only ghost & surface (rhsdmat)
              call set_rhsdmat()
          end do
        end if

!        if (s /= 0) then
!          dir = 1
!          delhead = surf_head(s) - head_new(i)
!          ! -- Set rhs & dmat for only ghost & surface (rhsdmat)
!            call set_rhsdmat()
!        end if

      else
        rhs(i) = DZERO
        dmat(i) = DONE
      end if
    end do

    return

  end subroutine form_boundterm

  subroutine set_rhsdmat()
  !*****************************************************************************
  ! set_rhsdmat -- Set rhs & dmat for only ghost & surface
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: deri_gs
    !---------------------------------------------------------------------------
    if (dir >= 4) then
      ori = 7 - dir
    else
      ori = dir
    end if

    abyd = area(c,dir)/dis(c,dir)
    rhs(i) = rhs(i) - hydcon_new(i,ori)*abyd*delhead*rel_perm(i)
    deri_gs = -hydcon_new(i,ori)*abyd*rel_perm(i)

    if (form_swich == 1 .and. sim_flag >= 0) then
      deri_kphi1 = (pertur_rel(i)-rel_perm(i))*newper_inv
      deri_gs = deri_gs + deri_kphi1*delhead*abyd
    end if

    dmat(i) = dmat(i) + deri_gs

    return

  end subroutine set_rhsdmat

end module make_linearsystem
