module calc_function
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE
  use read_inputfile, only: sim_flag
  use set_classification, only: nunknow, calc2cell, calc2cals, cals2surf, calc_end
  use make_cell, only: area, vol_c, dis
  use set_condition, only: off_index, off_row, conn_dir
  use set_calculation, only: bound_cflag, rech_cflag, well_cflag, cell_ss,&
                             cell_sat, surf_head
  use calc_parameter, only: calc_relperm, calc_hyd_upwind, calc_hyd_harm
  use allocate_module, only: head_old, head_new, srati_old, srati_new,&
                             hydcon_new, calc_rech, calc_well, rel_perm
  use time_module, only: delt

  implicit none
  private
  public :: calc_func, calc_vecjacf

  ! -- local
  integer(I4) :: i, j, k, s, c, dir
  real(DP) :: conn_flow, delhead, abyd, delt_inv

  contains

  subroutine calc_func(infx, funcv)
  !*****************************************************************************
  ! calc_func -- Calculate function value
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: infx(:)
    real(DP), intent(out) :: funcv(nunknow)
    ! -- local

    !---------------------------------------------------------------------------
    funcv(:) = DZERO
    ! -- Calculate relative permeability (relperm)
      call calc_relperm(nunknow, DZERO, head_new, srati_new, rel_perm)

    if (sim_flag >= 0) then
      delt_inv = DONE/delt
      ! -- Function storage change (stochn)
        call func_stochn(infx, funcv)
    end if

    ! -- Function connect flow from adjacent cells (connflow)
      call func_connflow(infx, funcv)

    ! -- Function boundary term (boundterm)
      call func_boundterm(infx, funcv)

    return

  end subroutine calc_func

  subroutine func_stochn(infstoc, outfstoc)
  !*****************************************************************************
  ! func_stochn -- Function storage change
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: infstoc(:)
    real(DP), intent(inout) :: outfstoc(:)
    ! -- local
    real(DP) :: deri_stor, ch_stor1, ch_stor2, ch_stor, alp
    !---------------------------------------------------------------------------
    do i = 1, nunknow
      deri_stor = DZERO
      if (bound_cflag(i) /= 1) then
        c = calc2cell(i)
        ch_stor1 = srati_old(i)*head_old(i)
        ch_stor2 = srati_new(i)*infstoc(i)
        if (srati_new(i) < DONE) then
          alp = DZERO
        else
          alp = DONE
        end if
        ch_stor = (alp*cell_ss(i)*(ch_stor2-ch_stor1) + &
                  cell_sat(i)*(srati_new(i)-srati_old(i)))*vol_c(c)

        outfstoc(i) = outfstoc(i) - ch_stor*delt_inv
      end if
    end do

    return

  end subroutine func_stochn

  subroutine func_connflow(infconn, outfconn)
  !*****************************************************************************
  ! func_connflow -- Function connect flow from adjacent cells
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: infconn(:)
    real(DP), intent(inout) :: outfconn(:)
    ! -- local
    real(DP) :: dis1, dis2, dis12, conn_hyd
    real(DP) :: hyd_face, hyd_face1, hyd_face2, relat
    integer(I4) :: ori
    !---------------------------------------------------------------------------
    do i = 1, nunknow
      conn_flow = DZERO
      c = calc2cell(i)
      do k = off_index(i-1)+1, off_index(i)
        dir = conn_dir(k)
        dis1 = dis(c,dir)
        j = off_row(k)
        delhead = infconn(j) - infconn(i)

        if (bound_cflag(j) == 1) then
          dis2 = dis1
          dis12 = dis1
          j = i
        else
          dis2 = dis(calc2cell(j),(7-dir))
          dis12 = dis1 + dis2
        end if
        abyd = area(c,dir)/dis12

        if (dir >= 4) then
          ori = 7 - dir
        else
          ori = dir
        end if

        ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
          call calc_hyd_upwind(infconn(i), infconn(j), rel_perm(i), rel_perm(j), relat)

        hyd_face1 = hydcon_new(i,ori)
        hyd_face2 = hydcon_new(j,ori)

        ! -- Calculate hydradulic conductivity by harmonic mean (hyd_harm)
          call calc_hyd_harm(hyd_face1, hyd_face2, dis1, dis2, hyd_face)

        conn_hyd = hyd_face*relat

        conn_flow = conn_flow + conn_hyd*abyd*delhead

      end do
      outfconn(i) = outfconn(i) + conn_flow
    end do

    return

  end subroutine func_connflow

  subroutine func_boundterm(infboun, outfboun)
  !*****************************************************************************
  ! func_boundterm -- Function boundary term
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: infboun(:)
    real(DP), intent(inout) :: outfboun(:)
    ! -- local
    integer(I4) :: ss
    !---------------------------------------------------------------------------
    do i = 1, nunknow
      if (bound_cflag(i) /= 1) then
        s = calc2cals(i)
        c = calc2cell(i)
        ! neuman boundary condition & well uptake
        if (s /= 0) then
          if (rech_cflag(s) == 1) then
            outfboun(i) = outfboun(i) + calc_rech(s)*area(c,1)
          end if
        end if
        if (well_cflag(i) == 1) then
          outfboun(i) = outfboun(i) + calc_well(i)
        end if

        if (s /= 0) then
          ss = cals2surf(s)
            dir = 1
            delhead = surf_head(s) - infboun(i)
            abyd = area(c,dir)/dis(c,dir)
            conn_flow = hydcon_new(i,dir)*abyd*delhead*rel_perm(i)
            outfboun(i) = outfboun(i) + conn_flow
        end if
      else
        outfboun(i) = DZERO
      end if
    end do

    return

  end subroutine func_boundterm

  subroutine calc_vecjacf(vjlevel, injx, injvec, outjvec)
  !*****************************************************************************
  ! calc_vecjacf -- Calculate vector by jacobi-free
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: vjlevel
    real(DP), intent(in) :: injx(:), injvec(:)
    real(DP), intent(out) :: outjvec(:)
    ! -- local
    integer(I4) :: vj_num
    real(DP) :: eps, eps_inv, l2_x, l2_v, sign, machine_eps = 1.0E-06
    real(DP), allocatable :: jcvec(:), tempf1(:), tempf2(:)
    !---------------------------------------------------------------------------
    vj_num = calc_end(vjlevel) - calc_end(vjlevel-1)

    allocate(jcvec(vj_num), tempf1(vj_num), tempf2(vj_num))
    jcvec(:) = DZERO ; tempf1(:) = DZERO ; tempf2(:) = DZERO

    ! -- Calculate function value (func)
      call calc_func(injx, tempf1)

      ! NITSOL version
  !    ! -- Calculate Calculate l2 norm square
  !      call calc_l2norm2(vjlevel, tempf1, l2_x)
  !    ! -- Calculate Calculate l2 norm square
  !      call calc_l2norm2(vjlevel, injvec, l2_v)

  !    eps = sqrt((DONE+sqrt(l2_x))*machine_eps)/sqrt(l2_v)
      ! Brown and Saad version
      l2_v = DZERO ; l2_x = DZERO
      do i = 1, vj_num
        l2_v = l2_v + injvec(i)*injvec(i)
        l2_x = l2_x + injx(i)*injvec(i)
      end do
      if (l2_x < DZERO) then
        sign = -DONE
      else
        sign = DONE
      end if
      eps = sign*machine_eps*max(abs(l2_x),sqrt(l2_v))/l2_v

      ! simple version
  !    l2_v = DZERO ; l2_x = DZERO
  !    do i = 1, vj_num
  !      l2_v = l2_v + injvec(i)*injvec(i)
  !      l2_x = l2_x + abs(injx(i))*machine_eps + machine_eps
  !    end do
  !    eps = l2_x/(l2_v*vj_num)

      eps_inv = DONE/eps

    do i = 1, vj_num
      jcvec(i) = injx(i) + eps*injvec(i)
    end do

    ! -- Calculate function value (func)
      call calc_func(jcvec, tempf2)

    do i = 1, vj_num
      outjvec(i) = -(tempf2(i)-tempf1(i))*eps_inv
    end do

    deallocate(jcvec, tempf1, tempf2)

    return

  end subroutine calc_vecjacf

end module calc_function
