module calc_massbalance
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DONE
  use read_inputfile, only: sim_flag
  use set_classification, only: calc2cell, calc2cals, cals2surf
  use make_cell, only: area, vol_c, dis
  use set_condition, only: off_index, off_row, conn_dir, g_dir, g_index
  use set_calculation, only: bound_cflag, rech_cflag, well_cflag, cell_ss,&
                             cell_sat, surf_head, ghost_head
  use calc_parameter, only: calc_relperm, calc_hyd_upwind, calc_hyd_harm
  use allocate_module, only: hydcon_new, calc_rech, calc_well, rel_perm
  use time_module, only: delt

  implicit none
  private
  public :: calc_mass

  ! -- local
  integer(I4) :: i, j, k, c, s, dir, ori
  real(DP) :: conn_flow, delhead, abyd, conn_hyd

  contains

  subroutine calc_mass(s_flag, cnum, h_new, h_pre, re_ms, we_ms, st_ms, co_ms)
  !*****************************************************************************
  ! calc_mass -- Calculate massbalance
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: s_flag, cnum
    real(DP), intent(in) :: h_new(:), h_pre(:)
    real(DP), intent(out) :: re_ms(cnum), we_ms(cnum)
    real(DP), intent(out) :: st_ms(cnum), co_ms(cnum)
    ! -- local
    real(DP), allocatable :: s_new(:), s_pre(:)
    real(DP) :: stor0, stor1, ch_srati, dis1, dis2, dis12, alp0, alp1
    integer(I4) :: g, ss
    real(DP) :: hyd_face, hyd_face1, hyd_face2, relat
    !---------------------------------------------------------------------------
    allocate(s_new(cnum), s_pre(cnum))
    s_new(:) = DZERO ; s_pre(:) = DZERO

    call calc_relperm(cnum, DZERO, h_pre, s_pre, rel_perm)
    call calc_relperm(cnum, DZERO, h_new, s_new, rel_perm)

    do i = 1, cnum
      if (bound_cflag(i) /= 1) then
        s = calc2cals(i)
        c = calc2cell(i)
        ! calculate mass from neumann boundary
        if (s /= 0) then
          ss = cals2surf(s)
          if (rech_cflag(s) == 1) then
            re_ms(s) = re_ms(s) + calc_rech(s)*area(c,1)*delt
          end if
        end if
        if (well_cflag(i) == 1) then
          we_ms(i) = we_ms(i) + calc_well(i)*delt
        end if

        ! calculate mass in storage
        if (sim_flag == -1) then
          st_ms(i) = DZERO
        else
          if (s_pre(i) < DONE) then
            alp0 = DZERO
          else if (s_new(i) < DONE) then
            alp1 = DZERO
          else
            alp0 = DONE ; alp1 = DONE
          end if
          stor0 = alp0*cell_ss(i)*s_pre(i)*h_pre(i)
          stor1 = alp1*cell_ss(i)*s_new(i)*h_new(i)
          ch_srati = cell_sat(i)*(s_new(i)-s_pre(i))
          st_ms(i) = st_ms(i) - (stor1-stor0+ch_srati)*vol_c(c)
        end if

        conn_flow = DZERO

        ! dirichlet bounday condtion except sea level (surface water & MG ghost)
        if (bound_cflag(i) == 2) then !MG ghost
          do g = g_index(i-1)+1, g_index(i)
            dir = g_dir(g)
            delhead = ghost_head(g) - h_new(i)
            ! -- Calculate connect flow for only ghost & surface (conf)
              call calc_conf()
          end do
        end if

        if (s /= 0 .and. s_flag == 0) then
          dir = 1
          delhead = surf_head(s) - h_new(i)
          ! -- Calculate connect flow for only ghost & surface (conf)
            call calc_conf()
        end if

        ! calculate mass between cells
        do k = off_index(i-1)+1, off_index(i)
          dir = conn_dir(k)
          dis1 = dis(c,dir)
          j = off_row(k)
          delhead = h_new(j) - h_new(i)

          if (bound_cflag(j) == 1) then
            dis2 = dis1
            dis12 = dis1
            j = i
          else
            dis2 = dis(calc2cell(j),(7-dir))
            dis12 = dis1 + dis2
          end if
          abyd = area(c,dir)/dis12

          if (dir >= 4) then
            ori = 7 - dir
          else
            ori = dir
          end if

          ! -- Calculate hydradulic conductivity by upwind (hyd_upwind)
            call calc_hyd_upwind(h_new(i), h_new(j), rel_perm(i), rel_perm(j), relat)

          hyd_face1 = hydcon_new(i,ori)
          hyd_face2 = hydcon_new(j,ori)

          ! -- Calculate hydradulic conductivity by harmonic mean (hyd_harm)
            call calc_hyd_harm(hyd_face1, hyd_face2, dis1, dis2, hyd_face)

          conn_hyd = hyd_face*relat

          conn_flow = conn_flow + conn_hyd*abyd*delhead

        end do
        co_ms(i) = co_ms(i) + conn_flow*delt
      end if
    end do

    return

  end subroutine calc_mass

  subroutine calc_conf()
  !*****************************************************************************
  ! calc_conf -- Calculate connect flow for only ghost & surface
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    if (dir >= 4) then
      ori = 7 - dir
    else
      ori = dir
    end if

    abyd = area(c,dir)/dis(c,dir)
    conn_flow = conn_flow + hydcon_new(i,ori)*abyd*delhead*rel_perm(i)

    return

  end subroutine calc_conf

end module calc_massbalance
