module make_cell
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DHALF, FACE, DQUART
  use read_inputfile, only: nx, ny, nz, xpoint, ypoint, zpoint, nlevel
  use set_classification, only: ncell, nsurf, nxgrid, nygrid, nzgrid, c_end, s_end

  implicit none
  private
  public :: make_cellinfo

  real(DP), allocatable, public :: dis(:,:), area(:,:)
  real(DP), allocatable, public :: vol_c(:), surfelev(:)
  real(DP), allocatable, public :: cell_topelev(:), cell_centelev(:), cell_botelev(:)

  ! -- local
  integer(I4) :: i, j, k, ii, jj, kk, n, c_num
  real(DP), save, allocatable :: fp_xw(:,:,:), fp_yw(:,:,:), fp_xs(:,:,:), fp_ys(:,:,:)
  real(DP), save, allocatable :: fp_xe(:,:,:), fp_ye(:,:,:), fp_xn(:,:,:), fp_yn(:,:,:)
  real(DP), save, allocatable :: cp_x(:,:,:), cp_y(:,:,:)
  real(DP), save, allocatable :: fp_zw(:,:,:,:), fp_zs(:,:,:,:), fp_ze(:,:,:,:), fp_zn(:,:,:,:)
  real(DP), save, allocatable :: fp_zt(:,:,:,:), fp_zb(:,:,:,:), cp_z(:,:,:,:)
  real(DP), allocatable :: fp_xtb(:,:,:), fp_ytb(:,:,:)
  integer(I4), allocatable :: skip(:)

  contains

  subroutine make_cellinfo()
  !*****************************************************************************
  ! make_cellinfo -- Make cell information
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Make center point on each face & cell (centpoint)
      call make_centpoint()

    ! -- Make distance between point and face center (disp2face)
      call make_disp2face()

    ! -- Make face area (facearea)
      call make_facearea()

    ! -- Make cell volume (cellvol)
      call make_cellvol()

    ! -- Make cell surface, top, center elevation (cellstcelev)
      call make_cellstcelev()

    return

  end subroutine make_cellinfo

  subroutine make_centpoint()
  !*****************************************************************************
  ! make_centpoint -- Make center point information
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(fp_xw(nlevel,nx,ny), fp_yw(nlevel,nx,ny), fp_xs(nlevel,nx,ny), fp_ys(nlevel,nx,ny))
    allocate(fp_xe(nlevel,nx,ny), fp_ye(nlevel,nx,ny), fp_xn(nlevel,nx,ny), fp_yn(nlevel,nx,ny))
    allocate(fp_xtb(nlevel,nx,ny), fp_ytb(nlevel,nx,ny), cp_x(nlevel,nx,ny), cp_y(nlevel,nx,ny))
    allocate(fp_zw(nlevel,nx,ny,nz), fp_zs(nlevel,nx,ny,nz), fp_ze(nlevel,nx,ny,nz), fp_zn(nlevel,nx,ny,nz))
    allocate(fp_zt(nlevel,nx,ny,nz), fp_zb(nlevel,nx,ny,nz), cp_z(nlevel,nx,ny,nz))
    allocate(skip(nlevel))
    fp_xw(:,:,:) = DZERO ; fp_yw(:,:,:) = DZERO ; fp_xs(:,:,:) = DZERO ; fp_ys(:,:,:) = DZERO
    fp_xe(:,:,:) = DZERO ; fp_ye(:,:,:) = DZERO ; fp_xn(:,:,:) = DZERO ; fp_yn(:,:,:) = DZERO
    fp_xtb(:,:,:) = DZERO ; fp_ytb(:,:,:) = DZERO ; cp_x(:,:,:) = DZERO ; cp_y(:,:,:) = DZERO
    fp_zw(:,:,:,:) = DZERO ; fp_zs(:,:,:,:) = DZERO ; fp_ze(:,:,:,:) = DZERO ; fp_zn(:,:,:,:) = DZERO
    fp_zt(:,:,:,:) = DZERO ; fp_zb(:,:,:,:) = DZERO ; cp_z(:,:,:,:) = DZERO
    skip(:) = 1

    ! make center point for face & cell
    do n = 1, nlevel
      skip(n) = 2**(n-1)
      do jj = 1, nygrid(n)
        j = (jj-1)*skip(n) + 1
        do ii = 1, nxgrid(n)
          i = (ii-1)*skip(n) + 1
          fp_xw(n,ii,jj) = DHALF*(xpoint(i,j) + xpoint(i,j+skip(n)))
          fp_yw(n,ii,jj) = DHALF*(ypoint(i,j) + ypoint(i,j+skip(n)))
          fp_xs(n,ii,jj) = DHALF*(xpoint(i,j) + xpoint(i+skip(n),j))
          fp_ys(n,ii,jj) = DHALF*(ypoint(i,j) + ypoint(i+skip(n),j))
          fp_xtb(n,ii,jj) = DQUART*(xpoint(i,j) + xpoint(i,j+skip(n)) + &
                            xpoint(i+skip(n),j) + xpoint(i+skip(n),j+skip(n)))
          fp_ytb(n,ii,jj) = DQUART*(ypoint(i,j) + ypoint(i,j+skip(n)) + &
                            ypoint(i+skip(n),j) + ypoint(i+skip(n),j+skip(n)))
        if (ii /= 1) then
          fp_xe(n,ii-1,jj) = fp_xw(n,ii,jj)
          fp_ye(n,ii-1,jj) = fp_yw(n,ii,jj)
        end if
        if (jj /= 1) then
          fp_xn(n,ii,jj-1) = fp_xs(n,ii,jj)
          fp_yn(n,ii,jj-1) = fp_ys(n,ii,jj)
        end if
        if (ii == nxgrid(n)) then
          fp_xe(n,ii,jj) = DHALF*(xpoint(i+skip(n),j) + xpoint(i+skip(n),j+skip(n)))
          fp_ye(n,ii,jj) = DHALF*(ypoint(i+skip(n),j) + ypoint(i+skip(n),j+skip(n)))
        end if
        if (jj == nygrid(n)) then
          fp_xn(n,ii,jj) = DHALF*(xpoint(i,j+skip(n)) + xpoint(i+skip(n),j+skip(n)))
          fp_yn(n,ii,jj) = DHALF*(ypoint(i,j+skip(n)) + ypoint(i+skip(n),j+skip(n)))
        end if

        do kk = 1, nzgrid(n)
          k = (kk-1)*skip(n) + 1
          fp_zw(n,ii,jj,kk) = DQUART*(zpoint(i,j,k) + zpoint(i,j,k+skip(n)) &
                            + zpoint(i,j+skip(n),k) + zpoint(i,j+skip(n),k+skip(n)))
          fp_zs(n,ii,jj,kk) = DQUART*(zpoint(i,j,k) + zpoint(i,j,k+skip(n)) &
                            + zpoint(i+skip(n),j,k) + zpoint(i+skip(n),j,k+skip(n)))
          fp_zt(n,ii,jj,kk) = DQUART*(zpoint(i,j,k) + zpoint(i,j+skip(n),k) &
                            + zpoint(i+skip(n),j,k) + zpoint(i+skip(n),j+skip(n),k))

          if (ii /= 1) then
            fp_ze(n,ii-1,jj,kk) = fp_zw(n,ii,jj,kk)
          end if
          if (jj /= 1) then
            fp_zn(n,ii,jj-1,kk) = fp_zs(n,ii,jj,kk)
          end if
          if (kk /= 1) then
            fp_zb(n,ii,jj,kk-1) = fp_zt(n,ii,jj,kk)
          end if
          if (ii == nxgrid(n)) then
            fp_ze(n,ii,jj,kk) = DQUART*(zpoint(i+skip(n),j,k) + zpoint(i+skip(n),j+skip(n),k) &
                         + zpoint(i+skip(n),j,k+skip(n)) + zpoint(i+skip(n),j+skip(n),k+skip(n)))
          end if
          if (jj == nygrid(n)) then
            fp_zn(n,ii,jj,kk) = DQUART*(zpoint(i,j+skip(n),k) + zpoint(i+skip(n),j+skip(n),k) &
                         + zpoint(i,j+skip(n),k+skip(n)) + zpoint(i+skip(n),j+skip(n),k+skip(n)))
          end if
          if (kk == nzgrid(n)) then
            fp_zb(n,ii,jj,kk) = DQUART*(zpoint(i,j,k+skip(n)) + zpoint(i+skip(n),j,k+skip(n)) &
                         + zpoint(i+skip(n),j,k+skip(n)) + zpoint(i+skip(n),j+skip(n),k+skip(n)))
          end if
        end do
      end do
      end do
    end do

    do n = 1, nlevel
      do ii = 1, nxgrid(n)
        do jj = 1, nygrid(n)
          do kk = 1, nzgrid(n)
            cp_z(n,ii,jj,kk) = DHALF*(fp_zt(n,ii,jj,kk) + fp_zb(n,ii,jj,kk))
          end do
          cp_x(n,ii,jj) = fp_xtb(n,ii,jj)
          cp_y(n,ii,jj) = fp_ytb(n,ii,jj)
        end do
      end do
    end do

    deallocate(fp_xtb, fp_ytb)

    return

  end subroutine make_centpoint

  subroutine make_disp2face()
  !*****************************************************************************
  ! make_disp2face -- Make distance between point and face center
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(dis(ncell,FACE))
    dis(:,:) = DZERO

    ! make distance between point and face center
    do n = 1, nlevel
      do k = 1, nzgrid(n)
        do j = 1, nygrid(n)
          do i = 1, nxgrid(n)
            c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + i
            dis(c_num,1) = fp_zt(n,i,j,k) - cp_z(n,i,j,k)
            dis(c_num,6) = cp_z(n,i,j,k) - fp_zb(n,i,j,k)
            dis(c_num,3) = distance(cp_x(n,i,j), fp_xw(n,i,j), cp_y(n,i,j),&
                           fp_yw(n,i,j), cp_z(n,i,j,k), fp_zw(n,i,j,k))
            dis(c_num,4) = distance(cp_x(n,i,j), fp_xe(n,i,j), cp_y(n,i,j),&
                           fp_ye(n,i,j), cp_z(n,i,j,k), fp_ze(n,i,j,k))
            dis(c_num,2) = distance(cp_x(n,i,j), fp_xs(n,i,j), cp_y(n,i,j),&
                           fp_ys(n,i,j), cp_z(n,i,j,k), fp_zs(n,i,j,k))
            dis(c_num,5) = distance(cp_x(n,i,j), fp_xn(n,i,j), cp_y(n,i,j),&
                           fp_yn(n,i,j), cp_z(n,i,j,k), fp_zn(n,i,j,k))
          end do
        end do
      end do
    end do

    deallocate(fp_xw, fp_yw, fp_xs, fp_ys, fp_xe, fp_ye, fp_xn, fp_yn)
    deallocate(fp_zw, fp_ze, fp_zs, fp_zn)

    return

  end subroutine make_disp2face

  function distance(x1, x2, y1, y2, z1, z2) result(d)
  !*****************************************************************************
  ! distance -- Calculate distance between two points
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: x1, x2, y1, y2, z1, z2
    ! -- local
    real(DP) :: d
    !---------------------------------------------------------------------------
    d = (x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2
    d = sqrt(d)

  end function distance

  subroutine make_facearea()
  !*****************************************************************************
  ! meke_facearea -- Make face area
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(area(ncell,FACE))
    area(:,:) = DZERO

    ! make face area
    do n = 1, nlevel
      skip(n) = 2**(n-1)
      do kk = 1, nzgrid(n)
        k = (kk-1)*skip(n) + 1
        do jj = 1, nygrid(n)
          j = (jj-1)*skip(n) + 1
          do ii = 1, nxgrid(n)
            i = (ii-1)*skip(n) + 1
            c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(kk-1) + nxgrid(n)*(jj-1) + ii

            area(c_num,1) = farea(xpoint(i,j), xpoint(i+skip(n),j),&
                          xpoint(i+skip(n),j+skip(n)), xpoint(i,j+skip(n)), ypoint(i,j),&
                          ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                          ypoint(i,j+skip(n)), zpoint(i,j,k),&
                          zpoint(i+skip(n),j,k), zpoint(i+skip(n),j+skip(n),k), zpoint(i,j+skip(n),k))

            area(c_num,6) = farea(xpoint(i,j), xpoint(i+skip(n),j),&
                          xpoint(i+skip(n),j+skip(n)), xpoint(i,j+skip(n)), ypoint(i,j),&
                          ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                          ypoint(i,j+skip(n)), zpoint(i,j,k+skip(n)),&
                          zpoint(i+skip(n),j,k+skip(n)), zpoint(i+skip(n),j+skip(n),k+skip(n)), zpoint(i,j+skip(n),k+skip(n)))

            area(c_num,3) = farea(xpoint(i,j), xpoint(i,j+skip(n)),&
                          xpoint(i,j+skip(n)), xpoint(i,j), ypoint(i,j),&
                          ypoint(i,j+skip(n)), ypoint(i,j+skip(n)), ypoint(i,j),&
                          zpoint(i,j,k), zpoint(i,j+skip(n),k),&
                          zpoint(i,j+skip(n),k+skip(n)), zpoint(i,j,k+skip(n)))

            area(c_num,4) = farea(xpoint(i+skip(n),j), xpoint(i+skip(n),j+skip(n)),&
                          xpoint(i+skip(n),j+skip(n)), xpoint(i+skip(n),j),&
                          ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                          ypoint(i+skip(n),j+skip(n)),ypoint(i+skip(n),j),&
                          zpoint(i+skip(n),j,k), zpoint(i+skip(n),j+skip(n),k),&
                          zpoint(i+skip(n),j+skip(n),k+skip(n)), zpoint(i+skip(n),j,k+skip(n)))

            area(c_num,2) = farea(xpoint(i,j), xpoint(i,j),&
                          xpoint(i+skip(n),j), xpoint(i+skip(n),j), ypoint(i,j),&
                          ypoint(i,j), ypoint(i+skip(n),j), ypoint(i+skip(n),j),&
                          zpoint(i,j,k), zpoint(i,j,k+skip(n)),&
                          zpoint(i+skip(n),j,k+skip(n)), zpoint(i+skip(n),j,k))

            area(c_num,5) = farea(xpoint(i,j+skip(n)), xpoint(i+skip(n),j+skip(n)),&
                          xpoint(i+skip(n),j+skip(n)), xpoint(i,j+skip(n)),&
                          ypoint(i,j+skip(n)), ypoint(i+skip(n),j+skip(n)),&
                          ypoint(i+skip(n),j+skip(n)), ypoint(i,j+skip(n)),&
                          zpoint(i,j+skip(n),k), zpoint(i+skip(n),j+skip(n),k),&
                          zpoint(i+skip(n),j+skip(n),k+skip(n)), zpoint(i,j+skip(n),k+skip(n)))
          end do
        end do
      end do
    end do

    return

  end subroutine make_facearea

  function farea(x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4) result(triarea)
  !*****************************************************************************
  ! area -- Calculate face area using cross product with daigonal vector
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4
    ! -- local
    real(DP) :: triarea
    real(DP) :: vec3d1(3), vec3d2(3)
    !---------------------------------------------------------------------------
    vec3d1(1) = x3 - x1 ; vec3d1(2) = y3 - y1 ; vec3d1(3) = z3 - z1
    vec3d2(1) = x4 - x2 ; vec3d2(2) = y4 - y2 ; vec3d2(3) = z4 - z2
    triarea = (vec3d1(1)*vec3d2(2) - vec3d1(2)*vec3d2(1))**2 + &
              (vec3d1(2)*vec3d2(3) - vec3d1(3)*vec3d2(2))**2 + &
              (vec3d1(3)*vec3d2(1) - vec3d1(1)*vec3d2(3))**2

    triarea = DHALF*sqrt(triarea)

  end function farea

  subroutine make_cellvol()
  !*****************************************************************************
  ! meke_cellvol -- Make cell volume
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(vol_c(ncell))
    vol_c(:) = DZERO

    ! make cell volume
    do n = 1, nlevel
      skip(n) = 2**(n-1)
      do kk = 1, nzgrid(n)
        k = (kk-1)*skip(n) + 1
        do jj = 1, nygrid(n)
          j = (jj-1)*skip(n) + 1
          do ii = 1, nxgrid(n)
            i = (ii-1)*skip(n) + 1
            c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(kk-1) + nxgrid(n)*(jj-1) + ii
            ! west side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j),&
                         xpoint(i,j+skip(n)), xpoint(i,j+skip(n)), cp_y(n,ii,jj),&
                         ypoint(i,j), ypoint(i,j+skip(n)), ypoint(i,j+skip(n)),&
                         cp_z(n,ii,jj,kk),zpoint(i,j,k), zpoint(i,j+skip(n),k),&
                         zpoint(i,j+skip(n),k+skip(n)))
            ! west side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j+skip(n)),&
                         xpoint(i,j), xpoint(i,j), cp_y(n,ii,jj),&
                         ypoint(i,j+skip(n)), ypoint(i,j), ypoint(i,j),&
                         cp_z(n,ii,jj,kk), zpoint(i,j+skip(n),k+skip(n)),&
                         zpoint(i,j,k+skip(n)), zpoint(i,j,k))
            ! east side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j),&
                         xpoint(i+skip(n),j), xpoint(i+skip(n),j+1), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j), ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j,k), zpoint(i+skip(n),j,k+skip(n)),&
                         zpoint(i+skip(n),j+skip(n),k+skip(n)))
            ! east side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j+skip(n)),&
                         xpoint(i+skip(n),j+skip(n)), xpoint(i+skip(n),j), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j+skip(n)), ypoint(i+skip(n),j+skip(n)), ypoint(i+skip(n),j),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j+skip(n),k+skip(n)),&
                         zpoint(i+skip(n),j+skip(n),k), zpoint(i+skip(n),j,k))
            ! south side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j),&
                         xpoint(i,j), xpoint(i+skip(n),j), cp_y(n,ii,jj),&
                         ypoint(i,j), ypoint(i,j), ypoint(i+skip(n),j),&
                         cp_z(n,ii,jj,kk), zpoint(i,j,k), zpoint(i,j,k+skip(n)),&
                         zpoint(i+skip(n),j,k+skip(n)))
            ! soutn side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j),&
                         xpoint(i+skip(n),j), xpoint(i,j), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j), ypoint(i+skip(n),j), ypoint(i,j),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j,k+skip(n)), zpoint(i+skip(n),j,k), zpoint(i,j,k))
            ! north side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j+skip(n)),&
                         xpoint(i+skip(n),j+skip(n)), xpoint(i+skip(n),j+skip(n)), cp_y(n,ii,jj),&
                         ypoint(i,j+skip(n)), ypoint(i+skip(n),j+skip(n)), ypoint(i+skip(n),j+skip(n)),&
                         cp_z(n,ii,jj,kk), zpoint(i,j+skip(n),k),&
                        zpoint(i+skip(n),j+skip(n),k), zpoint(i+skip(n),j+skip(n),k+skip(n)))
            ! north side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j+skip(n)),&
                         xpoint(i,j+skip(n)), xpoint(i,j+skip(n)), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j+skip(n)), ypoint(i,j+skip(n)), ypoint(i,j+skip(n)),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j+skip(n),k+skip(n)),&
                         zpoint(i,j+skip(n),k+skip(n)), zpoint(i,j+skip(n),k))
            ! top side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j),&
                         xpoint(i+skip(n),j), xpoint(i+skip(n),j+skip(n)), cp_y(n,ii,jj),&
                         ypoint(i,j), ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                         cp_z(n,ii,jj,kk), zpoint(i,j,k), zpoint(i+skip(n),j,k), zpoint(i+skip(n),j+skip(n),k))
            ! top side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j+skip(n)),&
                         xpoint(i,j+skip(n)), xpoint(i,j), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j+skip(n)), ypoint(i,j+skip(n)), ypoint(i,j),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j+skip(n),k), zpoint(i,j+skip(n),k), zpoint(i,j,k))
            ! bottom side1
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i,j),&
                         xpoint(i+skip(n),j), xpoint(i+skip(n),j+skip(n)), cp_y(n,ii,jj),&
                         ypoint(i,j), ypoint(i+skip(n),j), ypoint(i+skip(n),j+skip(n)),&
                         cp_z(n,ii,jj,kk), zpoint(i,j,k+skip(n)), zpoint(i+skip(n),j,k+skip(n)),&
                         zpoint(i+skip(n),j+skip(n),k+skip(n)))
            ! bottom side2
            vol_c(c_num) = vol_c(c_num) + vol(cp_x(n,ii,jj), xpoint(i+skip(n),j+skip(n)),&
                         xpoint(i,j+skip(n)), xpoint(i,j), cp_y(n,ii,jj),&
                         ypoint(i+skip(n),j+skip(n)), ypoint(i,j+skip(n)), ypoint(i,j),&
                         cp_z(n,ii,jj,kk), zpoint(i+skip(n),j+skip(n),k+skip(n)),&
                         zpoint(i,j+skip(n),k+skip(n)), zpoint(i,j,k+skip(n)))
          end do
        end do
      end do
    end do

    deallocate(cp_x, cp_y)
    deallocate(cp_z)
    deallocate(skip)

    return

  end subroutine make_cellvol

  function vol(x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4) result(trivol)
  !*****************************************************************************
  ! vol -- Calculate cell vol using cross product with a focus on cell center
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4
    ! -- local
    real(DP) :: trivol
    real(DP) :: vec3d1(3), vec3d2(3), vec3d3(3), cropro12(3)
    !---------------------------------------------------------------------------
    vec3d1(1) = x2 - x1 ; vec3d1(2) = y2 - y1 ; vec3d1(3) = z2 - z1
    vec3d2(1) = x3 - x1 ; vec3d2(2) = y3 - y1 ; vec3d2(3) = z3 - z1
    vec3d3(1) = x4 - x1 ; vec3d3(2) = y4 - y1 ; vec3d3(3) = z4 - z1
    cropro12(1) = vec3d1(2)*vec3d2(3) - vec3d1(3)*vec3d2(2)
    cropro12(2) = vec3d1(1)*vec3d2(3) - vec3d1(3)*vec3d2(1)
    cropro12(3) = vec3d1(1)*vec3d2(2) - vec3d1(2)*vec3d2(1)

    trivol = dot_product(abs(cropro12(1:3)), abs(vec3d3(1:3)))
    trivol = abs(trivol)/(6.00_DP)

  end function vol

  subroutine make_cellstcelev()
  !*****************************************************************************
  ! make_cellstcelev -- Make cell surface, top, center elevation
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: s_num

    !---------------------------------------------------------------------------
    ! -- allocate
    allocate(cell_topelev(ncell), cell_centelev(ncell), cell_botelev(ncell))
    allocate(surfelev(nsurf))
    cell_topelev(:) = DZERO ; cell_centelev(:) = DZERO ; cell_botelev(:) = DZERO
    surfelev(:) = DZERO

    ! make cell top & bottom elevation
    do n = 1, nlevel
      do k = 1, nzgrid(n)
        if (k == 1) then
          do j = 1, nygrid(n)
            do i = 1, nxgrid(n)
              s_num = s_end(n-1) + nxgrid(n)*(j-1) + i
              c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + i
              surfelev(s_num) = fp_zt(n,i,j,k)
              cell_topelev(c_num) = fp_zt(n,i,j,k)
              cell_botelev(c_num) = fp_zb(n,i,j,k)
              cell_centelev(c_num) = DHALF*(fp_zt(n,i,j,k) + fp_zb(n,i,j,k))
            end do
          end do
        else
          do j = 1, nygrid(n)
            do i = 1, nxgrid(n)
              c_num = c_end(n-1) + nxgrid(n)*nygrid(n)*(k-1) + nxgrid(n)*(j-1) + i
              cell_topelev(c_num) = fp_zt(n,i,j,k)
              cell_botelev(c_num) = fp_zb(n,i,j,k)
              cell_centelev(c_num) = DHALF*(fp_zt(n,i,j,k) + fp_zb(n,i,j,k))
            end do
          end do
        end if
      end do
    end do

    deallocate(fp_zt, fp_zb)

    return

  end subroutine make_cellstcelev

end module make_cell
