module nonlinear_solution
  ! -- modules
  use kind_module, only: I4, DP
  use constvalue_module, only: DZERO, DHALF, DONE, DTWO, VARIABLELENGTH,&
                               VARMAX, XMAX
  use read_inputfile, only: maxout_iter, sim_flag, criteria, outinit_flag,&
                            nx, ny, calc_unit, outche_flag, errtol
  use set_classification, only: ncalc, ncals, nunknow, cals2calc, calc2cell,&
                                cals2surf
  use make_cell, only: surfelev, area
  use set_condition, only: off_index
  use set_calculation, only: bound_cflag, surf_cflag, cell_surf, surf_head,&
                             current_t, seep_cflag
  use allocate_module, only: rhs, dmat, lumat, head_new, head_pre, head_change
  use time_module, only: conv_flag, form_swich, delt, now_time
  use calc_massbalance, only: calc_mass
  use calc_simulation, only: calc_resl2norm, calc_l2norm2
  use calc_function, only: calc_func, calc_vecjacf
  use make_linearsystem, only: make_matvec
  use check_simulation, only: check_abserrmax
  use linear_solution, only: solve_linalg, iter

  implicit none
  private
  public :: calc_numsol

  ! -- local
  integer(I4) :: out_iter, back_iter, back_flag
  integer(I4) :: i, j, k, c
  character(len=VARIABLELENGTH) :: cxyzn
  real(DP) :: maxvar, maxunk, check_val, check_pre
  integer(I4) :: maxnun
  real(DP) :: l2_new, l2_pre, l2_init, l2_jnorm, lam, eta, gradient, maxstep
  real(DP), allocatable :: new_r(:)

  contains

  subroutine calc_numsol()
  !*****************************************************************************
  ! calc_numsol -- Calculate numerical solution
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    ! -- format
    10 format(//1x,"OUTER ITERATION : ",es12.6,1x,"(",a,")",5x,"TIME STEP : ",&
              es10.3,1x,"(SEC)",/,1x,78("-"),/,&
              1x," OUTER INNER  BACK     MAXIMUM     MAXIMUM      DIALOG  &
              RIGHT HAND     UNKNOWN",/,1x,&
              "                        CHANGE        CELL      MATRIX      &
              VECTOR       VALUE",/,1x,78("-"))
    11 format(1X,3(i6),es12.3,a12,4(es12.3))
    12 format(1X,"Didn't converge due to maximum value or change")
    13 format(1X,"Stop due to maximum value or change in backtracking")
    14 format(1X,"Stop due to maximum number of nonlinear iteration")
    !---------------------------------------------------------------------------
    outer_loop : do out_iter = 1, maxout_iter

      if (out_iter == 1) then
        write(101,10) now_time, trim(calc_unit), delt
        head_pre(1:nunknow) = head_new(1:nunknow)
        form_swich = 1 ; back_flag = 0 ; back_iter = 0
        allocate(new_r(nunknow))
        new_r(:) = DZERO
      else
        l2_pre = l2_new ; check_pre = check_val
      end if

      ! -- Reset coefficients matrix and constant vector (matvec)
        call reset_matvec()

      ! -- Calculate surface water level (surfw)
        call calc_surfw()

      ! -- Make coefficients matrix and constant vector (matvec)
        call make_matvec()

      ! -- Set for backtracking (backtr)
        call set_backtr()

      if (out_iter == 1) then
        ! -- Calculate residual &l2norm square (resl2norm)
          call calc_resl2norm(1, nunknow, head_new, rhs, l2_init)
!          l2_pre = l2_init
          l2_pre = l2_init ; eta = DHALF
      end if

      head_pre(1:nunknow) = head_new(1:nunknow)
      head_change(1:nunknow) = DZERO

      if (sim_flag /= -1) then
        errtol = eta
      end if

      conv_flag = 0
      ! -- Solve linear algebra (linalg)
        call solve_linalg(head_change, l2_pre)

      do i = 1, nunknow
        head_new(i) = head_pre(i) + head_change(i)
      end do

      ! -- Calculate outer check value (checkv)
        call calc_checkv()

      if (conv_flag /= 0 .and. maxunk == maxunk) then
        conv_flag = 0
        if (check_val <= criteria .and. abs(maxunk) < XMAX) then
          conv_flag = 1
        end if
      else if (maxunk /= maxunk .and. sim_flag /= -1) then
        write(*,'(a)') "Nan detected stop calculation"
        exit outer_loop
      else if (maxunk /= maxunk .and. sim_flag == -1) then
        sim_flag = 0
        current_t = DZERO
        write(*,'(a)') "Steady state calculation changes to timestepcalculation"
        write(101,12)
        exit outer_loop
      end if

      if (out_iter /= 1 .and. form_swich == 1 .and. sim_flag /= -1) then
        ! -- Run backtracking (backtr)
          call run_backtr()
        ! -- Calculate outer check value (checkv)
          call calc_checkv()
        if (abs(maxvar) >= VARMAX .or. abs(maxunk) >= XMAX) then
          back_flag = 1
        else if (check_val <= criteria .and. abs(maxunk) < XMAX) then
          conv_flag = 1
        end if
        if (conv_flag /= 1) then
          ! -- Set eater (eater)
            call set_eater()
        end if
      end if

      if (conv_flag /= 1 .and. form_swich == 1 .and. sim_flag /= -1) then
!      ! -- Run under relax using delta-bar-delta underrelaxation (relaxdelta)
!        call run_relaxdelta(head_new, head_pre)
      ! -- Run under relax using cooley underrelaxation (relaxcooly)
        call run_relaxcooly(head_new, head_pre)

        do i = 1, nunknow
          head_change(i) = head_new(i) - head_pre(i)
        end do

        ! -- Calculate outer check value (checkv)
          call calc_checkv()

        if (check_val <= criteria .and. abs(maxunk) < XMAX) then
          conv_flag = 1
        else if (back_flag == 1 .and. abs(maxvar) < VARMAX .and. abs(maxunk) < XMAX) then
          back_flag = 0
        end if
      end if

      ! -- Calculate function value (func)
        call calc_func(head_new, new_r)
      ! -- Calculate Calculate l2 norm square
        call calc_l2norm2(1, new_r, l2_new)

      if (out_iter*2 >= maxout_iter) then
        form_swich = 1
      end if

      ! -- Get cell number (cnum)
        call get_cnum(maxnun, cxyzn)
      write(101,11) out_iter, iter, back_iter, maxvar, trim(adjustl(cxyzn)),&
                    dmat(maxnun), rhs(maxnun), head_new(maxnun)
      ! check outer_loop
      if (conv_flag == 1) then
        if (outinit_flag > 0) then
          rewind(104)
          write(104) now_time, (head_new(i), i = 1, nunknow)
        end if
        exit outer_loop
      else if (back_flag == 1) then
        write(101,13)
        exit outer_loop
      else if (out_iter == maxout_iter) then
        write(101,14)
      else if (out_iter /= 1 .and. abs(maxvar) >= VARMAX .or. abs(maxunk) >= XMAX) then
        write(101,12)
        exit outer_loop
      end if

    end do outer_loop

    deallocate(new_r)

    return

  end subroutine calc_numsol

  subroutine reset_matvec
  !*****************************************************************************
  ! reset_matvec -- Reset coefficients matrix and constant vector
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    do i = 1, ncalc
      dmat(i) = DZERO
      rhs(i) = DZERO
      do j = off_index(i-1)+1, off_index(i)
        lumat(j) = DZERO
      end do
    end do

    return

  end subroutine reset_matvec

  subroutine calc_surfw()
  !*****************************************************************************
  ! calc_surfw -- Calculate surface water level
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    integer(I4) :: s, cc
    real(DP) :: sr_ms(ncals), sw_ms(ncals), ss_ms(ncals), sc_ms(ncals)
    real(DP) :: surf_ms
    !---------------------------------------------------------------------------
!    ! -- Calculate massbalance (mass)
!      call calc_mass(1, ncals, head_new, head_pre, sr_ms, sw_ms, ss_ms, sc_ms)

    do i = 1, ncals
      c = cals2calc(i)
      s = cals2surf(i)
      cc = calc2cell(c)
      if (surf_cflag(i) == 1) then
        surf_head(i) = cell_surf(i)
      else if (seep_cflag(i) == 1 .and. head_new(c) > surfelev(s)) then
        surf_head(i) = surfelev(s)
!      else if (surf_head(i) > surfelev(s) .and. bound_cflag(c) /= 1) then
!        surf_ms = sr_ms(i) + sw_ms(i) + ss_ms(i) + sc_ms(i)
!        surf_head(i) = surf_head(i) + surf_ms/area(cc,1)
      else
        surf_head(i) = head_new(c)
      end if
    end do

    return

  end subroutine calc_surfw

  subroutine set_backtr
  !*****************************************************************************
  ! set_backtr -- Set for backtracking
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: l2_xnew
    !---------------------------------------------------------------------------
    ! -- Calculate Calculate l2 norm square
      call calc_l2norm2(1, head_new, l2_xnew)
    maxstep = VARMAX*l2_xnew
    if (maxstep < DONE) then
      maxstep = DONE
    end if

    return

  end subroutine set_backtr

  subroutine set_eater
  !*****************************************************************************
  ! set_eater -- Set eater
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: eta_max = 0.9_DP, eta_min = 1.0E-4_DP, eta_safe
    real(DP) :: eta_alpha = (DONE+sqrt(5.0_DP))/DTWO, lin_l2norm, redfac2
    !---------------------------------------------------------------------------
    eta_safe = eta**eta_alpha
    lin_l2norm = sqrt(l2_pre + DTWO*lam*gradient + lam*l2_jnorm)
    eta = abs(sqrt(l2_new) - lin_l2norm)/sqrt(l2_pre)

    if (eta_safe < 0.1_DP) then
      eta_safe = DZERO
    end if

    eta = max(eta, eta_safe)
    eta = max(eta, eta_min)
    eta = min(eta, eta_max)

    return

  end subroutine set_eater

  subroutine run_backtr
  !*****************************************************************************
  ! run_backtr -- Run backtracking
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local
    real(DP) :: l2_new2, l2_pnorm, slope, av, bv, rhs1, rhs2, root, steplength
    real(DP) :: lam2, temp_lam, lam_inv, lam2_inv, del_lam, lam_max, lam_min
    real(DP) :: lam_length, lam_base, lam_diff, lam_incr, alpha_cond, beta_cond
    real(DP) :: back_alpha = 1.00E-4_DP, back_beta = 0.9_DP
    real(DP), allocatable :: jacvec(:)
    integer(I4) :: beta_iter = 0
    !---------------------------------------------------------------------------
    l2_new = DZERO ; l2_new2 = DZERO ; slope = DZERO ; lam = DONE ; lam2 = DZERO
    gradient = DZERO ; l2_pnorm = DZERO ; lam_max = DZERO ; lam_length = DZERO
    allocate(jacvec(nunknow))
    jacvec(:) = DZERO
    ! -- Calculate function value (func)
      call calc_func(head_new, new_r)
    ! -- Calculate Calculate l2 norm square
      call calc_l2norm2(1, head_change, l2_pnorm)
    ! -- Calculate vector by jacobi-free (vecjocf)
      call calc_vecjacf(1, head_pre, head_change, jacvec)
    ! -- Calculate Calculate l2 norm square
      call calc_l2norm2(1, jacvec, l2_jnorm)

    if (sqrt(l2_pnorm) > maxstep) then
      do i = 1, nunknow
        head_change(i) = head_change(i)*maxstep/sqrt(l2_pnorm)
      end do
      l2_pnorm = maxstep**DTWO
    end if
    steplength = sqrt(l2_pnorm)

    do i = 1, nunknow
      l2_new = l2_new + new_r(i)*new_r(i)
!      slope = slope + new_r(i)*jacvec(i)
      slope = slope + new_r(i)*(-jacvec(i))
      if (head_new(i) /= DZERO) then
        temp_lam = abs(head_change(i))/abs(head_new(i))
        if (temp_lam > lam_length) then
          lam_length = temp_lam
        end if
      end if
    end do

    back_iter = 0 ; lam_min = (back_alpha**DTWO)/lam_length; gradient = slope
    alpha_cond = DHALF*l2_pre + back_alpha*lam*slope
    back_aloop : do
      if (DHALF*l2_new <= alpha_cond) then
        exit back_aloop
      end if
      back_iter = back_iter + 1
      ! -- Calculate function & l2norm2 (func2norm)
        call calc_funcl2norm(lam)

      if (lam == DONE) then
        temp_lam = -slope/(DTWO*(DHALF*l2_new-DHALF*l2_pre-slope))
      else
        rhs1 = DHALF*l2_new - DHALF*l2_pre - lam*slope
        rhs2 = l2_new2 - DHALF*l2_pre - lam*slope
        lam_inv = DONE/(lam**2) ; lam2_inv = DONE/(lam2**2)
        del_lam = DONE/(lam - lam2)
        av = (rhs1*lam_inv - rhs2*lam2_inv)*del_lam
        bv = (-lam2*rhs1*lam_inv + lam*rhs2*lam2_inv)*del_lam
        if (av == 0) then
          temp_lam = -slope/(DTWO*bv)
        else
          root = bv*bv - 3.0_DP*av*slope
          if (root < DZERO) then
            temp_lam = DHALF*lam
          else if (bv <= DZERO) then
            temp_lam = (-bv + sqrt(root))/(3.0_DP*av)
          else
            temp_lam = -slope/(bv + sqrt(root))
          end if
          if (temp_lam > DHALF*lam) then
            temp_lam = DHALF*lam
          end if
        end if
      end if
      lam2 = lam
      l2_new2 = DHALF*l2_new
      lam = max(temp_lam, 0.1_DP*lam)
      if (lam < lam_min) then
        return
      end if
      alpha_cond = DHALF*l2_pre + back_alpha*lam*slope
    end do back_aloop

    alpha_cond = DHALF*l2_pre + back_alpha*lam*slope
    beta_cond = DHALF*l2_pre + back_beta*lam*slope
    if (DHALF*l2_new < beta_cond) then
      if (lam == DONE .and. sqrt(l2_pnorm) < steplength) then
        lam_max = steplength/sqrt(l2_pnorm)
        b1loop : do while (DHALF*l2_new <= alpha_cond .and. &
                           DHALF*l2_new < beta_cond .and. lam < lam_max)
          lam2 = lam ; lam = min(DTWO*lam, lam_max)
          ! -- Calculate function & l2norm2 (func2norm)
            call calc_funcl2norm(lam)
          alpha_cond = DHALF*l2_pre + back_alpha*lam*slope
          beta_cond = DHALF*l2_pre + back_beta*lam*slope
        end do b1loop
      end if

      if (lam < DONE .or. (lam > DONE .and. DHALF*l2_new > alpha_cond)) then
        lam_base = min(lam, lam2) ; lam_diff = abs(lam2 - lam)
        b2loop : do while (DHALF*l2_new > alpha_cond .or. &
                          (DHALF*l2_new < beta_cond .and. lam_diff > lam_min))
          lam_incr = DHALF*lam_diff ; lam = lam_base + lam_incr
          ! -- Calculate function & l2norm2 (func2norm)
            call calc_funcl2norm(lam)
          alpha_cond = DHALF*l2_pre + back_alpha*lam*slope
          beta_cond = DHALF*l2_pre + back_beta*lam*slope

          if (DHALF*l2_new > alpha_cond) then
            lam_diff = lam_incr
          else if (DHALF*l2_new < beta_cond) then
            lam_base = lam ; lam_diff = lam_diff - lam_incr
          end if
          if (lam_diff == DZERO) then
            exit b2loop
          end if
        end do b2loop

        if (DHALF*l2_new < beta_cond) then
          ! -- Calculate function & l2norm2 (func2norm)
            call calc_funcl2norm(lam_base)
          beta_iter = beta_iter + 1
        end if
        if (beta_iter == 10) then
          back_flag = 1
          return
        end if
      end if
    end if

    return

  end subroutine run_backtr

  subroutine run_relaxcooly(x_new, x_pre)
  !*****************************************************************************
  ! run_relaxcooly -- Run under relax using cooley underrelaxation
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: x_new(:), x_pre(:)
    ! -- local
    real(DP) :: relax = DONE, old_relax = DONE, old_maxvar = DZERO
    real(DP) :: esk = DZERO, esk_abs = DZERO, del_x = DZERO, max_del_x = DZERO
    !---------------------------------------------------------------------------
    if (out_iter == 1) then
      relax = DONE
      old_relax = DONE
      old_maxvar = maxvar
    else
      esk = maxvar/(old_maxvar*old_relax)
      esk_abs = abs(esk)
      if (esk < -DONE) then
        relax = DHALF/esk_abs
      else
        relax = (DONE*3 + esk)/(DONE*3 + esk_abs)
      end if
    end if

    old_relax = relax
    old_maxvar = maxvar

    if (relax < DONE) then
      do i = 1, nunknow
        if (bound_cflag(i) /= 1) then
          del_x = x_new(i) - x_pre(i)
          x_new(i) = x_pre(i) + relax*del_x
        end if
      end do
    end if

    return

  end subroutine run_relaxcooly

  subroutine run_relaxdelta(x_new, x_pre)
  !*****************************************************************************
  ! run_relaxdelta -- Run under relax using delta-bar-delta underrelaxation
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(inout) :: x_new(:), x_pre(:)
    ! -- local
    real(DP) :: old_w(nunknow), weight_x(nunknow), old_del_x(nunknow)
    real(DP) :: w, mom, theta = 0.97_DP, kappa = 1.0E-5_DP
    real(DP) :: gamma = DZERO, momentum = 0.1_DP, del_x
    !---------------------------------------------------------------------------
    do i = 1, nunknow
      if (bound_cflag(i) /= 1) then
        del_x = x_new(i) - x_pre(i)
        if (out_iter == 1) then
          old_w(i) = DONE
          old_del_x(i) = DZERO
          weight_x(i) = DZERO
        end if

        w = old_w(i)

        if (old_del_x(i)*del_x < DZERO) then
          w = theta*old_w(i)
        else
          w = old_w(i) + kappa
        end if

        if (w > DONE) then
          w = DONE
        end if

        old_w(i) = w

        if (out_iter == 1) then
          weight_x(i) = del_x
        else
          weight_x(i) = (DONE - gamma)*del_x + gamma*weight_x(i)
        end if

        old_del_x(i) = del_x

        mom = DZERO
        if (out_iter > 4) then
          mom = momentum
        end if
        del_x = del_x*w + mom*weight_x(i)
        x_new(i) = x_pre(i) + del_x
      end if
    end do

    return

  end subroutine run_relaxdelta

  subroutine calc_checkv()
  !*****************************************************************************
  ! calc_checkv -- Calculate outer check value
  !*****************************************************************************
    ! -- modules

    ! -- inout

    ! -- local

    !---------------------------------------------------------------------------
    ! -- Check absolute error max norm
      call check_abserrmax(head_new, head_pre, maxvar, maxunk, maxnun)

    if (outche_flag == 1) then
      check_val = abs(maxvar)
    else if (outche_flag == 2) then
      ! -- Calculate function value (func)
        call calc_func(head_new, new_r)
      ! -- Calculate Calculate l2 norm square
        call calc_l2norm2(1, new_r, l2_new)
      check_val = sqrt(l2_new)/sqrt(l2_init)
    end if

    return

  end subroutine calc_checkv

  subroutine get_cnum(calcnum, char_cell)
  !*****************************************************************************
  ! get_cnum -- Get cell number
  !*****************************************************************************
    ! -- modules

    ! -- inout
    integer(I4), intent(in) :: calcnum
    character(len=*), intent(out) :: char_cell
    ! -- local
    integer(I4) :: cell_num, xynum, xnum, ynum, znum
    character(len=12) :: cxnum, cynum, cznum
    !---------------------------------------------------------------------------
    cell_num = calc2cell(calcnum)
    znum = (cell_num-1)/(nx*ny) + 1
    xynum = cell_num - nx*ny*(znum-1)
    ynum = (xynum-1)/nx + 1
    xnum = xynum - (ynum-1)*nx

    write(cxnum,*) xnum ; write(cynum,*) ynum ; write(cznum,*) znum

    char_cell = "(" // trim(adjustl(cxnum)) // "," // trim(adjustl(cynum)) //&
                "," // trim(adjustl(cznum)) // ")"

    return

  end subroutine get_cnum

  subroutine calc_funcl2norm(in_lam)
  !*****************************************************************************
  ! calc_funcl2norm -- Calculate function & l2norm2
  !*****************************************************************************
    ! -- modules

    ! -- inout
    real(DP), intent(in) :: in_lam
    ! -- local

    !---------------------------------------------------------------------------
    do i = 1, nunknow
      head_new(i) = head_pre(i) + in_lam*head_change(i)
    end do

    ! -- Calculate function value (func)
      call calc_func(head_new, new_r)
    ! -- Calculate Calculate l2 norm square
      call calc_l2norm2(1, new_r, l2_new)

    return

  end subroutine calc_funcl2norm

end module nonlinear_solution
